#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#

# This is a translation to python of the ACSIS IfDisplayMainWindow.cc code
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import random
import traceback
import numpy

from qwt.qt.QtGui import (QApplication, QMainWindow, QDialog, QGridLayout,QHBoxLayout,QToolButton,
         QLabel, QSizePolicy, QSlider, QPushButton, QVBoxLayout, QSpinBox, QSpacerItem, QTabWidget, QPixmap)
from qwt.qt.QtGui import QPen, QColor,QWidget, QImage, qRgba, QFont, QFontInfo, QMenu, QActionGroup, QAction, QButtonGroup
from qwt.qt.QtCore import Qt, QSize, QObject, pyqtSignal

from MeqGUI.Plugins.chartplot_qt5 import ChartPlot

class ControlMenu (QMenu):
  """This is the control menu common to all the ChartPlot widgets"""
  changeComplexComponent = pyqtSignal(str)
  changeStokesComponent = pyqtSignal(str)
  changeVellsComponent = pyqtSignal(str)

  AMP = "Amplitude";
  PHASE = "Phase";
  REAL = "Real";
  IMAG = "Imaginary";
  STOKES_I = 'I'
  STOKES_Q = 'Q'
  STOKES_U = 'U'
  STOKES_V = 'V'
  ComplexComponents = (AMP,PHASE,REAL,IMAG);
  StokesComponents = (STOKES_I,STOKES_Q,STOKES_U,STOKES_V)
  ComplexComponentLabels = {AMP:"ampl",PHASE:"ph",REAL:"re",IMAG:"im"};
  StokesComponentLabels = {STOKES_I:"I",STOKES_Q:"Q",STOKES_U:"U",STOKES_V:"V"};
  def __init__ (self,parent):
    QMenu.__init__(self,parent);
    self.close_window = self.addAction('Close Window');
    self.close_window.setVisible(False)

    self.reset_zoomer = self.addAction('Reset zoomer');
    self.reset_zoomer.setVisible(False)

# the following is commented out until postscript/pdf printing works 
# properly with Qt 4 widgets
#   self._print = self.addAction("Print",self.plotPrinter.do_print)

    self.clear_plot = self.addAction('Clear plot');

    self.close_popups = self.addAction('Close popups')
    self.close_popups.setVisible(False)

    self.show_flagged_data = self.addAction('Show flagged data')
    self.show_flagged_data.setCheckable(True)
    self.show_flagged_data.setChecked(False)

    self.autoscale = self.addAction('Automatic scaling');
    self.autoscale.setCheckable(True)
    self.autoscale.setChecked(False)
    self.autoscale.setVisible(False) # not sure why, but this is never set to visible -- perhaps fixed scaling doesn't work at all

    self.offset_value = self.addAction('Offset Value');
    self.offset_value.setVisible(False); # not sure why, but this is never set to visible -- perhaps it no longer works at all
    
    self.show_labels = self.addAction('Show labels');
    self.show_labels.setCheckable(True)
    self.show_labels.setChecked(True)

    self.append = self.addAction('Accumulate data tracks')
    self.append.setCheckable(True)
    self.append.setChecked(True)

    # create submenu for complex data
    self.complex_menu = self.addMenu("Plot complex values as...");
    qag = QActionGroup(self.complex_menu);
    self._qas_complex = dict();
    self._tbs_complex = dict();
    for label in self.ComplexComponents:
      qa = self.complex_menu.addAction(label);
      qa.setCheckable(True);
      qag.addAction(qa);
      self._qas_complex[label] = qa;
    self._qas_complex[self.AMP].setChecked(True); 
    qag.triggered[QAction].connect(self._change_complex)
    self.complex_menu.menuAction().setVisible(False);
    # current complex component
    self.complex_component = self.AMP;

    # create submenu for correlated data element selection
    self.vells_menu = self.addMenu('Data element selector...')
    # menu and action group will be filled when the first updateEvent occurs
    self._qag_vells = QActionGroup(self.vells_menu);
    self._qas_vells = dict();
    self._tbs_vells = dict();
    self._qag_vells.triggered[QAction].connect(self._change_vells)
    self.vells_menu.menuAction().setVisible(False);
    self.vells_component = None;

    # create submenu for stokes data selection
    self.stokes_menu = self.addMenu('Select Stokes parameter...')
    # menu and action group will be filled when the first updateEvent occurs
    self._qag_stokes = QActionGroup(self.stokes_menu);
    self._qas_stokes = dict();
    self._tbs_stokes = dict();
    self._qag_stokes.triggered[QAction].connect(self._change_stokes)
    self.stokes_menu.menuAction().setVisible(False);
    self.stokes_component = None;

    self.save_this = self.addAction('Save this plot page in PNG format');
    self.save_all = self.addAction('Save all pages in PNG format');

  def createDataSelectorWidgets (self,parent,parent_layout):
    """Creates toolbuttons for complex values and Vells selection""";
    
    #print('in createDataSelectionWidgets')
    self._ds_top = top = QWidget(parent);
    parent_layout.addWidget(top);
    self._ds_lo = lotop = QVBoxLayout(top);
    lotop.setContentsMargins(0,0,0,0);
    self._ds_complex = QWidget(top);
    self._ds_complex.setVisible(False);
    lotop.addWidget(self._ds_complex);
    lo = QVBoxLayout(self._ds_complex);
    lo.setContentsMargins(0,0,0,0);
    lab = QLabel("complex:");
    lab.setAlignment(Qt.AlignHCenter);
    lo.addWidget(lab);
    # add complex selector
    lo0 = QHBoxLayout();
    lo0.setContentsMargins(0,0,0,0);
    lo.addLayout(lo0);
    lo1 = QGridLayout()
    lo1.setContentsMargins(0,0,0,0);
    lo1.setHorizontalSpacing(0);
    lo1.setVerticalSpacing(0);
#   lo0.addStretch(1);
    lo0.addLayout(lo1);
#   lo0.addStretch(1);
    bgrp = QButtonGroup(self._ds_complex);
#   tbdesc = { self.AMP:(u"\u007Ca\u007C",0,0),self.PHASE:(u"\u03D5",0,1),self.REAL:("Re",1,0),self.IMAG:("Im",1,1) };
#   tbdesc = { self.AMP:("\\u007Ca\\u007C",0,0),self.PHASE:("\\u0278",0,1),self.REAL:("Re",1,0),self.IMAG:("Im",1,1) };
    tbdesc = { self.AMP:("Amp",0,0),self.PHASE:("Pha",0,1),self.REAL:("Re",1,0),self.IMAG:("Im",1,1) };
    for label,qa in list(self._qas_complex.items()):
      tbtext,row,col = tbdesc[label];
      tb = QToolButton(self._ds_complex);
      lo1.addWidget(tb,row,col);
      bgrp.addButton(tb);
      tb.setText(tbtext);
      tb.setToolButtonStyle(Qt.ToolButtonTextOnly);
      tb.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Minimum);
      tb.setCheckable(True);
      tb.setChecked(label is self.complex_component);
      tb.setMinimumWidth(32);
      tb.clicked[bool].connect(qa.setChecked)
      tb.clicked[bool].connect(self._change_complex)
      qa.triggered[bool].connect(tb.setChecked)
      self._tbs_complex[label] = tb;


  def setVellsElementLabels(self,labels,dims):
    # do nothing when only one label, or when already set
    if len(labels)<2 or self._qas_vells:
      return;
    # make menu items
#   print 'in setVellsElementLabels, labels = ', labels
    for label in labels:
      # make menu action
      self._qas_vells[label] = va = self._qag_vells.addAction(str(label));
      va.setCheckable(True);
      # if first QAction, then check it
      if len(self._qas_vells) == 1:
        va.setChecked(True)
        self.vells_component = label;
      self.vells_menu.addAction(va);
    self.vells_menu.menuAction().setVisible(True);

# following does nothing at the moment
    for label in self.StokesComponents:
      self._qas_stokes[label] = vs = self._qag_stokes.addAction(label);
      vs.setCheckable(True);
      self.stokes_menu.addAction(vs);
    self.stokes_menu.menuAction().setVisible(True);
    # make grid of selector buttons, if dims are not too big

    if len(dims) == 1:
      dims = (1,dims[0]);
    if len(dims) == 2 and min(*dims)>=2 and max(*dims)<=6:
      # for dims=1, make it 1xN 
      # add vells selector 
      self._ds_lo.addSpacing(16);
      self._ds_vells = QWidget(self._ds_top);
      self._ds_lo.addWidget(self._ds_vells);
      self._ds_stokes = QWidget(self._ds_top);
      self._ds_lo.addWidget(self._ds_stokes);
      self._ds_stokes.setVisible(False);
      lo = QVBoxLayout(self._ds_vells);
      lo.setContentsMargins(0,0,0,0);
      lab = QLabel("element:");
      lab.setAlignment(Qt.AlignHCenter);
      lo.addWidget(lab);
      # add data selectors for correlations and Stokes
      lo0 = QVBoxLayout();
      lo0.setContentsMargins(0,0,0,0);
      lo.addLayout(lo0);
      lo1 = QGridLayout()
      lo1.setContentsMargins(0,0,0,0);
      lo1.setHorizontalSpacing(0);
      lo1.setVerticalSpacing(0);
      lo0.addLayout(lo1);
      bgrp = QButtonGroup(self._ds_vells);
      # make the labels
      for ilabel,label in enumerate(labels):
        # make toolbutton
        tb = QToolButton(self._ds_vells);
        bgrp.addButton(tb);
        self._tbs_vells[label] = tb;
        tb.setText(str(label));
        tb.setToolButtonStyle(Qt.ToolButtonTextOnly);
        tb.setCheckable(True);
        tb.setChecked(label is self.vells_component);
        tb.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Minimum);
  #      tb.setMinimumWidth(32);
        qa = self._qas_vells[label];
        tb.clicked[bool].connect(qa.setChecked)
        tb.clicked[bool].connect(self._change_vells)
        qa.triggered[bool].connect(tb.setChecked)
        # add to layout in correct place
        row,col = divmod(ilabel,dims[1]);
        if dims[1] > 3:
          col,row = row,col;
        lo1.addWidget(tb,row,col);
      # show/hide controls
      self._ds_vells.setVisible(len(labels) > 1);

      lab = QLabel("stokes:");
      lab.setAlignment(Qt.AlignHCenter);
      lo0.addWidget(lab);

      lo2 = QGridLayout()
      lo2.setContentsMargins(0,0,0,0);
      lo2.setHorizontalSpacing(0);
      lo2.setVerticalSpacing(0);
      lo0.addLayout(lo2);
      bgrp = QButtonGroup(self._ds_stokes);
      stdesc = {self.STOKES_I:("I",0,0),self.STOKES_Q:("Q",0,1),self.STOKES_U:("U",1,0),self.STOKES_V:("V",1,1) };
      for label,qa in list(self._qas_stokes.items()):
        tbtext,row,col = stdesc[label];
        tb = QToolButton(self._ds_stokes);
        lo2.addWidget(tb,row,col);
        bgrp.addButton(tb);
        tb.setText(tbtext);
        tb.setToolButtonStyle(Qt.ToolButtonTextOnly);
        tb.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Minimum);
        tb.setCheckable(True);
        tb.setChecked(label is self.stokes_component);
        qa = self._qas_stokes[label];
        tb.clicked[bool].connect(qa.setChecked)
        tb.clicked[bool].connect(self._change_stokes)
        qa.triggered[bool].connect(tb.setChecked)
        self._tbs_complex[label] = tb;
      # show/hide controls
      self._ds_stokes.setVisible(len(labels) > 1);

  def isComplexControlVisible (self):
    return self.complex_menu.menuAction().isVisible();
    
  def isVellsControlVisible (self):
    return self.vells_menu.menuAction().isVisible();

  def showComplexControls (self,show=True):
    """Enables complex controls. If called at least once, they become enabled and stay visible.""";
    if show:
      self.complex_menu.menuAction().setVisible(True);
      try:
        self._ds_complex and self._ds_complex.setVisible(True);
      except:
        pass

  def _change_complex (self,*dum):
    for label,qa in list(self._qas_complex.items()):
      if qa.isChecked():
        self.complex_component = label;
        break;
    self.autoscale.setChecked(True);
    self.changeComplexComponent.emit(self.complex_component)

  def _change_stokes (self,*dum):
    for label,qa in list(self._qas_stokes.items()):
      if qa.isChecked():
        self.stokes_component = label;
#       print(('setting Stokes to ', label))
        break;
    self.autoscale.setChecked(True);
    self.changeStokesComponent.emit(self.stokes_component)
    
  def _change_vells (self,*dum):
    for label,qa in list(self._qas_vells.items()):
      if qa.isChecked():
        self.vells_component = label;
        if label in self._tbs_vells:
          self._tbs_vells[label].setChecked(True);
          #print(('setting vells label to ', label))
        break;
    self.autoscale.setChecked(True);
    self.changeVellsComponent.emit(self.vells_component)

# def updateTabSelectorRange(self, max_range):
#   """ set or update maximum range for tab page selector """
#   print 'updatomg tab selector to ', max_range
#   self._tab_selector.set_emit(False)
#   self._tab_selector.setMinValue(0)
#   self._tab_selector.setMaxValue(max_range,False)
#   self._tab_selector.setValue(max_range)
#   self._tab_selector.set_emit(True)
#   self._tab_selector.show()


class DisplayMainWindow(QMainWindow):
  """ This class enables the display of a collection
      of ChartPlot widgets contained within a tabwidget
  """
  number_of_tabs = pyqtSignal(int)
  auto_offset_value = pyqtSignal(float)
  showMessage = pyqtSignal(str)

  def __init__(self, parent=None, name=None,num_curves=16,plot_label=None):
    QMainWindow.__init__(self, parent)

# ChartPlot strip charts will be displayed via a tab widget
    self._tabwidget = QTabWidget(self)
    self._tab_resized = False
    self._num_curves = num_curves
    self._plot_label = plot_label
    self._result_range = None
    self._png_number = 0
    self._grab_name = ''

# create control menu
    self._menu = ControlMenu(self);
# create a dictionary of chart plot objects
    self._ChartPlot = {}
    self._click_on = "If you click on an individual stripchart with the <b>middle</b> mouse button, a popup window will appear that gives a more detailed plot of the data from that particular object. <br><br> Clicking with the <b>left</b> mouse button will cause a small popup to appear. The popup gives the actual X and Y values, corrected for offset, of the data point nearest to the location of the mouse.<br><br> Clicking with the <b>right</b> mouse button will cause a context menu to appear. The <b>Accumulate data tracks</b> option means that data in each tile will be appended to the previous data. If this option is unchecked, data will be displayed for just each individual tile. The <b>Data element selector</b> option works similarly to that associated with the standard 2-D plot display. Clicking on it causes a small submenu to appear that allows you to select different data elements for display."

    # connect menu signals
    self._menu.save_this.triggered.connect(self.save_current_display)
    self._menu.save_all.triggered.connect(self.save_all_displays)

  def createDataSelectorWidgets (self,parent,layout):
    self._menu.createDataSelectorWidgets(parent,layout);

  def setDataElementLabels (self,labels,dims):
    self._menu.setVellsElementLabels(labels,dims);


  def setDataElementLabels (self,labels,dims):
#   print 'setDataElementLabels incoming labels,dims ', labels,dims
#   print labels
    self._menu.setVellsElementLabels(labels,dims);


  def updateEvent(self, data_dict):
    data_type = data_dict['data_type']
#   print('Display updating with type', data_type)
    try:
      self._grab_name = data_dict['source']
    except:
      self._grab_name = ''
#   print('Display updating with _grab_name', self._grab_name)
    if data_type not in self._ChartPlot:
      self._ChartPlot[data_type] = ChartPlot(self._menu,num_curves=self._num_curves,parent=self)
      self._ChartPlot[data_type].setDataLabel(data_type)
#     self._ChartPlot[data_type].set_plotter_inactive()
#     self._ChartPlot[data_type].set_plotter_active()
      self._max_tab_index = self._tabwidget.addTab(self._ChartPlot[data_type], data_type)
#     print 'tabwidget index ', self._max_tab_index
      self.number_of_tabs.emit(self._max_tab_index)
      
#     self._menu.updateTabSelectorRange(self._max_tab_index)
      self._tabwidget.setCurrentWidget(self._ChartPlot[data_type])
      self._tabwidget.resize(self._tabwidget.minimumSizeHint())
      self.resize(self._tabwidget.minimumSizeHint())

      dcm_sn_descriptor = "This window shows stripcharts of data as a function of time. The display is mostly used to show radio interferometer data where the frequency data have been averaged together. Each tab window can show up to 64 interferometer baselines. The antenna pair associated with each baseline is shown with a yellow background.<br><br>"
      dcm_sn_descriptor = dcm_sn_descriptor + self._click_on
      self._ChartPlot[data_type].setWhatsThis(dcm_sn_descriptor)
      try:
        self._ChartPlot[data_type].quit_event.connect(self.quit_event)
      except:
        pass
      self._ChartPlot[data_type].auto_offset_value.connect(self.report_auto_value)
      if not self._plot_label is None:
        self._ChartPlot[data_type].setPlotLabel(self._plot_label)
      self._ChartPlot[data_type].show()
      # make "Save all" visibile if multiple pages
      self._menu.save_all.setVisible(len(self._ChartPlot) > 1);
      
    self._ChartPlot[data_type].updateEvent(data_dict)
    self._ChartPlot[data_type].setSource(self._grab_name)

  def change_tab_page(self, tab_page):
    self._tabwidget.setCurrentIndex(tab_page)

  def report_auto_value(self, auto_offset_value):
    self.auto_offset_value.emit(auto_offset_value)

  def set_range_selector(self, new_range):
    """ set or update maximum range for slider controller """
    for plot in list(self._ChartPlot.values()):
      plot.set_offset_scale(new_range)

  def set_auto_scaling(self):
    """ set or update maximum range for slider controller """
    for plot in list(self._ChartPlot.values()):
      plot.set_auto_scaling()

  def resizeEvent(self, event):
    for plot in list(self._ChartPlot.values()):
      plot.resize(event.size())
    self._tabwidget.resize(event.size())

  def setNewPlot(self):
    for plot in list(self._ChartPlot.values()):
      plot.clear_plot()

  def save_current_display (self):
    filename,error = self._save_display(self._tabwidget.currentWidget());
    if error:
      self.showMessage.emit("error writing file %s"%filename, True)
    else:
      self.showMessage.emit("saved plot to %s"%filename)
    
  def save_all_displays (self):
    good_files = [];
    bad_files = [];
    for key in sorted(self._ChartPlot.keys()):
      filename,error = self._save_display(self._ChartPlot[key]);
      (bad_files if error else good_files).append(filename);
    if good_files:
      self.showMessage.emit("saved plots to %s"%(", ".join(good_files)))
    if bad_files:
      self.showMessage.emit("error writing files %s"%(", ".join(bad_files)), True)
    
  def _save_display (self,chartplot):
    self._png_number += 1;
    # put together filename components
    name_components = [];
    if self._grab_name:
      name_components.append(self._grab_name);
    if self._menu.isVellsControlVisible():
      name_components.append(str(self._menu.vells_component));
    if self._menu.isComplexControlVisible():
      name_components.append(str(self._menu.ComplexComponentLabels[self._menu.complex_component]));
    if len(self._ChartPlot) > 1:
      name_components.append(chartplot.dataLabel());
    name_components.append(str(self._png_number));
    save_file = "_".join(name_components).replace(' ','_')+".png";
    try:
      pm = QPixmap.grabWidget(chartplot);
      pm.save(save_file, "PNG");
      return save_file,None;
    except:
      traceback.print_exc();
      print('failed to grab or save pixmap')
      return save_file,True;
 


#void IfDisplayMainWindow.set_data_flag(Int channel, Bool flag_value)
#{
#	_ChartPlot(0).set_data_flag(channel,flag_value)
#
## if flag_value == False, we have bad data!
#	if (!flag_value) {
#    		QColor col
#		col.setNamedColor("IndianRed")
#    		statusBar().setPaletteBackgroundColor(col)
#		QString channel_number
#		channel_number.setNum(channel)
#		QString Message = "Bad TSYS detected for channel "+ channel_number 
#    		statusBar().message( Message)
#        	QTimer *timer = new QTimer(this)
#        	connect( timer, SIGNAL(timeout()), this, SLOT(resetStatus()) )
## TRUE means that this will be a one-shot timer
#        	timer.start(500, TRUE)
#	}
#}

#void IfDisplayMainWindow.resetStatus()
#{
#    		QColor col
#    		col.setNamedColor("LightYellow")
#    		statusBar().setPaletteBackgroundColor(col)
#		QString Message = " "
#    		statusBar().message( Message)
#}
#
  def quit_event(self):
    self.close()

  def start_test_timer(self, time):
    # stuff for tests
    self.seq_num = 0
    self._gain = 0
    self._array = numpy.zeros((128,), numpy.float32)
    self._array_imag = numpy.zeros((128,), numpy.float32)
    self._array_complex = numpy.zeros((128,), numpy.complex64)
    self.startTimer(time)

  def timerEvent(self, e):
    self.seq_num = self.seq_num + 1
    self._gain = self._gain + 0.5
    data_dict = {}
    data_dict['sequence_number'] = self.seq_num

    for i in range(16):
      data_dict['channel'] = i

      data_dict['data_type'] = 'scalar'
      data_dict['value'] = (i+1) * random.random()
      self.updateEvent(data_dict)

      data_dict['data_type'] = 'another scalar'
      data_dict['value'] = self._gain + (i+1) * random.random()
      self.updateEvent(data_dict)

      data_dict['data_type'] = 'arrays'
      if i == 13:
        for j in range(self._array.shape[0]):
          self._array[j] = 11 * random.random()
          self._array_imag[j] = 6 * random.random()
        self._array_complex.real = self._array
        self._array_complex.imag = self._array_imag
        data_dict['value'] = self._array_complex
      else:
        for j in range(self._array.shape[0]):
          self._array[j] = (i+1) * random.random()
        data_dict['value'] = self._array
      self.updateEvent(data_dict)

      data_dict['data_type'] = 'tensor demo'
      data_dict['value'] = {}
      for j in range(4):
        if j == 0 or j == 3:
          gain = 1.0
          for k in range(self._array.shape[0]):
            self._array[k] = gain * random.random()
          data_dict['value'][j] = self._array.copy()
        else:
          gain = 0.1
          for k in range(self._array.shape[0]):
            self._array[k] = gain * random.random()
            self._array_imag[k] = gain * random.random()
          self._array_complex.real = self._array
          self._array_complex.imag = self._array_imag
          data_dict['value'][j] = self._array_complex.copy()
      self.updateEvent(data_dict)

    return

def make():
    demo = DisplayMainWindow()
    demo.show()
    demo.start_test_timer(1000)
    return demo

def main(args):
    app = QApplication(args)
    demo = make()
    app.exec_()


# Admire
if __name__ == '__main__':
    main(sys.argv)




# -*- coding: utf-8 -*-
#
# Licensed under the terms of the PyQwt License
# Copyright (C) 2003-2009 Gerard Vermeulen, for the original PyQwt example
# Copyright (c) 2015 Pierre Raybaut, for the PyQt5/PySide port and further 
# developments (e.g. ported to PythonQwt API)
# (see LICENSE file for more details)

import sys
import numpy

from qwt.qt.QtGui import QApplication, QPen, QBrush
from qwt.qt.QtCore import QSize, QSizeF, QRect, QRectF, QPoint, QPointF
from qwt.qt.QtCore import Qt
from qwt import QwtPlot, QwtSymbol, QwtPlotGrid, QwtPlotCurve, QwtText

class QwtPlotCurveSizes(QwtPlotCurve):
    def __init__(self, x=[], y=[], symbolSizes=None):

        QwtPlotCurve.__init__(self)
        
        self.symbolSizes = symbolSizes
        self.symbolList = None
        self.setData(x, y, symbolSizes)

    def setData(self, *args):
        if len(args) == 1:
            QwtPlotCurve.setData(self, *args)
            return
        x, y = args[:2]
        if len(args) > 2:
            self.symbolSizes = args[2]
        
        self.__x = numpy.asarray(x, numpy.float)
        if len(self.__x.shape) != 1:
            raise RuntimeError('len(asarray(x).shape) != 1')
        self.__y = numpy.asarray(y, numpy.float)
        if len(self.__y.shape) != 1:
            raise RuntimeError('len(asarray(y).shape) != 1')
        if len(self.__x) != len(self.__y):
            raise RuntimeError('len(asarray(x)) != len(asarray(y))')
        QwtPlotCurve.setData(self, self.__x, self.__y)

    def setSymbolList(self, symbolList):
      """ Override default QwtCurve symbols """
      self.symbolList = symbolList

    def drawSymbols(self, painter, symbol, xMap, yMap, canvasRect,from_, to):
      if self.symbolList is None and self.symbolSizes is None:
        print('QwtPlotCurveSizes fail: you must specify a symbol list or an array of symbol sizes')
        return
      painter.setBrush(symbol.brush());
      painter.setPen(symbol.pen())
      for i in range(from_, to+1):
        if not self.symbolList is None:
          painter.setBrush(self.symbolList[i].brush());
          painter.setPen(self.symbolList[i].pen())
        xi = xMap.transform(self.__x[i]);
        yi = yMap.transform(self.__y[i]);
        position = QPointF(float(xi), float(yi))
        if not self.symbolList is None:
          self.symbolList[i].drawSymbol(painter, position);
        else:
          symbol.setSize(QSize(self.symbolSizes[i],self.symbolSizes[i]))
          symbol.drawSymbol(painter, position);
    # drawSymbols()

# class QwtPlotCurveSizes()

def make():
    # create a plot with a white canvas
    demo = QwtPlot(QwtText("Curve Demonstation"))
#   demo.setCanvasBackground(Qt.white)
#   demo.plotLayout().setAlignCanvasToScales(True)

#   grid = QwtPlotGrid()
#   grid.attach(demo)
#   grid.setPen(QPen(Qt.black, 0, Qt.DotLine))
    
    # calculate data and errors for a curve with error bars
    x = numpy.zeros(20, numpy.float32)
    y = numpy.zeros(20, numpy.float32)
    symbol_sizes = numpy.zeros(20, numpy.int32)
    symbolList=[]
    for i in range(20):
      x[i] = 1.0 * i
      y[i] = 2.0 * i
      symbol_sizes[i] = 3 + i
      if i%2 == 0:
        symbolList.append(QwtSymbol(QwtSymbol.UTriangle,
             QBrush(Qt.black), QPen(Qt.black), QSize(3+i,3+i)))
      else:
        symbolList.append(QwtSymbol(QwtSymbol.DTriangle,
             QBrush(Qt.red), QPen(Qt.red), QSize(3+i,3+i)))

    curve = QwtPlotCurveSizes(x = x, y = y, symbolSizes = symbol_sizes)
    x = x + 10
    curve1 = QwtPlotCurveSizes(x= x, y = y, symbolSizes = symbol_sizes)
    curve.setSymbol(QwtSymbol(QwtSymbol.Ellipse,
             QBrush(Qt.black), QPen(Qt.black), QSize(5,5)))
    curve.setPen(QPen(Qt.blue, 2))
    curve1.setSymbol(QwtSymbol(QwtSymbol.Ellipse,
             QBrush(Qt.red), QPen(Qt.red), QSize(10,10)))
    curve1.setPen(QPen(Qt.blue, 2))
    curve1.setSymbolList(symbolList)

    curve.attach(demo)
    curve1.attach(demo)
    demo.resize(640, 480)
    demo.show()
    return demo


if __name__ == '__main__':
    app = QApplication(sys.argv)
    demo = make()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import time

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

from Timba import dmi
from Timba.utils import PersistentCurrier

def hms_str (tm):
  """helper method, converts time in seconds to H:MM:SS string""";
  (tm,secs) = divmod(int(tm),60);
  (tm,mins) = divmod(tm,60);
  return "%d:%02d:%02d"%(tm,mins,secs);

class VisProgressMeter (QWidget):
  """VisProgressMeter implements a one-line progress meter
  to track progress messages from a VisDataMux. It is normally meant
  to be part of a status bar
  """;
  def __init__ (self,parent):
    QWidget.__init__(self,parent);
    lo = QHBoxLayout(self);
    lo.setContentsMargins(5,0,5,0);
    lo.setSpacing(5);
    self._wtime = QLabel(self);
    self._wtime.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum);
    self._wtime.setAlignment(Qt.AlignLeft|Qt.AlignVCenter);
    self._wtime.setIndent(5);
    self._wtime.setTextFormat(Qt.RichText);
    self._wprog = QProgressBar(self);
    self._wprog.setTextVisible(True);
    self._wprog.setMinimumWidth(128);
    self._wprog.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum);
    self._wlabel = QLabel(self);
    self._wlabel.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Minimum);
    self._wlabel.setIndent(5);
    self._wlabel.setTextFormat(Qt.RichText);
    self._wlabel.setAlignment(Qt.AlignLeft|Qt.AlignVCenter);
    self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Minimum);
    for w in (self._wtime,self._wprog,self._wlabel):
      lo.addWidget(w);
    self._app = None;
    self._vis_output = False;
    self._vis_inittime = None;
    self._timerid = None;
    self._currier = PersistentCurrier();
    self.curry = self._currier.curry;
    self.xcurry = self._currier.xcurry;
    self._timer_id = None;
    self._reset_stats();
    
  class TSStats (object):
    """helper class to contain secs/timeslot stats""";
    def __init__ (self):
      self.reset(None,0);
    def reset (self,tm,nt=None):
      """resets to time0, and optionally a number of timeslots""";
      self.time0 = tm;
      self.nt = nt or 0;
      self.rate = None;
    def mark (self,tm):
      """marks current time and computes rate""";
      if self.nt and self.time0 is not None:
        self.rate = (tm-self.time0)/self.nt;
      return self.rate;
    def update (self,nt):
      """adds number of timeslots""";
      self.nt += nt;
  
  def _reset_stats (self):
    self._stats = self.TSStats();  # total stats
    self._pstats = self.TSStats(); # previous tile's stats
    self._vis_nt = self._vis_rtime0 = self._vis_rtime1 = None;
    self._time_reset_needed = True;  # if true, total time counter will be reset
    self._hdr_time = self._etc_time = None;
    
  def connect_app_signals (self,app):
    """connects standard app signals to appropriate methods."""
    self._app = app;
    QObject.connect(app,PYSIGNAL("vis.channel.open"),self.xcurry(self.start,_argslice=slice(1,2)));
    QObject.connect(app,PYSIGNAL("vis.header"),self.xcurry(self.header,_argslice=slice(1,2)));
    QObject.connect(app,PYSIGNAL("vis.num.tiles"),self.xcurry(self.update,_argslice=slice(1,2)));
    QObject.connect(app,PYSIGNAL("vis.footer"),self.xcurry(self.footer,_argslice=slice(1,2)));
    QObject.connect(app,PYSIGNAL("vis.channel.closed"),self.xcurry(self.close,_argslice=slice(1,2)));
    QObject.connect(app,PYSIGNAL("isConnected()"),self.xcurry(self.reset));
    
  def _show (self):
    self._wlabel.show();
    self._wprog.show();
    self._wtime.show();
    self.show();
  
  def start (self,rec):
    """initializes and shows meter. Usually connected to a Vis.Channel.Open signal""";
    self._app.statusbar.clearMessage();
    self._wlabel.setText("<nobr>opening dataset</nobr>"); 
    self._wprog.reset();
    self._vis_output = rec.get("output");
    self._vis_inittime = time.time();
    self._wtime.setText(" 0:00:00");
    self._wtime.show();
    self._show();
    if self._timer_id is not None:
      self.killTimer(self._timer_id);
    self._timer_id = self.startTimer(1000);
    self._reset_stats();
    
  def timerEvent (self,ev):
    """redefined to keep clock display ticking"""
    self._update_time();
  
  def _update_time (self):
    txt = '';
    timestamp = time.time();
    if self._vis_inittime is not None:
      txt = hms_str(timestamp-self._vis_inittime);
      if self._etc_time is not None:
        txt += '<small> ETC </small>' + hms_str(self._etc_time-self._vis_inittime);
    self._wtime.setText("<nobr>"+txt+"</nobr>");
  
  def header (self,rec):
    """processes header record. Usually connected to a Vis.Header signal""";
    self._app.statusbar.clearMessage();
    times = rec.header.time_extent;
    nt = int(times[1] - times[0]);
    self._wprog.setMaximum(nt);
    if nt:
      timestr = "received header, dataset length is " + hms_str(nt);
    else:
      timestr = "received header";
    self._wlabel.setText("<nobr>"+timestr+"</nobr>"); 
    self._reset_stats();
    self._show();
    # init header time
    self._hdr_time = time.time();
       
  def update (self,rec):
    """indicates progress based on a Vis.Num.Tiles signal""";
    timestamp = time.time();
    if self._time_reset_needed:
      self._stats.reset(timestamp);
      self._time_reset_needed = False;
    nt = self._vis_nt = rec.num_tiles;
    ts = rec.timeslots;
    tm0 = rec.time_extent[0];
    nt0 = rec.time_extent[1]-rec.time_extent[0];
    time0 = int(rec.time[0]-tm0);
    time1 = int(rec.time[1]-tm0);
    self._wprog.setMaximum(nt0);
    self._wprog.setValue(time0);
    # compute ETC
    self._etc_time = None;
    if self._hdr_time is not None:
      progress = rec.time[0]-tm0;
      if progress:
        dt = timestamp - self._hdr_time;
        self._etc_time = self._hdr_time + dt*(nt0/progress);
    self._update_time();  
    # compute rates
    nts = ts[1]-ts[0]+1;
    self._stats.mark(timestamp);
    self._pstats.mark(timestamp);
    # form message
    timestr = self._vis_rtime1 = hms_str(time0);
    if self._vis_rtime0 is None:
      self._vis_rtime0 = timestr;
    if time1 != time0:
      timestr1 = self._vis_rtime1 = hms_str(time1);
      timestr += " to " + timestr1;
    msg = " tile <b>%d</b>, t/s %d-%d, r/t %s" \
      % (nt-1,ts[0],ts[1],timestr);
    if self._stats.rate is not None:
      msg = msg+"; avg <b>%.2f</b> sec/ts" % self._stats.rate;
#    if self._pstats.rate is not None:
#      msg = msg+"; last %.2f sec/ts" % self._pstats.rate;
    self._wlabel.setText("<nobr>"+msg+"</nobr>"); 
    # update stat counters
    self._stats.update(nts);
    self._pstats.reset(timestamp,nts);
    # show everything
    self._show();
    
  def footer (self,rec):
    """processes footer record. Usually connected to a Vis.Footer signal""";
    if self._vis_output:
      msg = "received footer, writing to output";
    else:
      msg = "received footer";
    timestamp = time.time();
    self._stats.mark(timestamp);
    self._pstats.mark(timestamp);
    if self._stats.rate is not None:
      msg = msg+"; avg <b>%.2f</b> sec/ts" % self._stats.rate;
    if self._pstats.rate is not None:
      msg = msg+"; last %.2f sec/ts" % self._pstats.rate;
    self._wlabel.setText("<nobr>"+msg+"</nobr>"); 
    self._wprog.setMaximum(100);
    self._wprog.setValue(99);
    self._show();
    
  def close (self,rec):
    """closes meter, posts message describing elapsed time and rate.
    Usually connected to a Vis.Channel.Closed signal.""";
    if self._app:
      msg = "dataset complete";
      rec = dmi.record();
      if self._vis_inittime is not None:
        elapsed = time.time()-self._vis_inittime;
        rec.elapsed = hms_str(elapsed);
        msg += " in "+rec.elapsed;
      else:
        elapsed = None;
      if self._stats.rate is not None:
        rec.secs_per_ts = self._stats.rate;
        msg = msg+"; avg %.2f sec/ts" % rec.secs_per_ts;
      if self._vis_nt is not None:
        rec.num_tiles = self._vis_nt;
        if elapsed is not None:
          rec.secs_per_tile = elapsed/self._vis_nt;
          msg = msg+"; avg %.2f sec/tile" % rec.secs_per_tile;
      if self._vis_rtime0 is not None:
        rec.start_time_rel = self._vis_rtime0;
      if self._vis_rtime1 is not None:
        rec.end_time_rel = self._vis_rtime1;
      self._app.log_message(msg,content=rec);
    self.reset();
    
  def reset (self):
    """resets and hides meter."""
    self._reset_stats();
    if self._timer_id is not None:
      self.killTimer(self._timer_id);
      self._timer_id = None;
    self._vis_inittime = None;
    self._wprog.reset();
    self._wprog.hide();
    self._wlabel.setText(""); 
    self._wlabel.hide();
    self._wtime.setText(""); 
    self._wtime.hide();
    self.hide();



#!/usr/bin/env python3

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from Timba.Meq import meqds
from Timba.Meq.meqds import mqs
from MeqGUI.GUI import browsers
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI.treebrowser import NodeAction
import MeqGUI.GUI.app_proxy_gui 
from MeqGUI.GUI.meqserver_gui import makeNodeDataItem
from MeqGUI import Grid
from Timba.array import *
from Timba.Meq import meq
from time import sleep
from qt import *
from string import *
from operator import isNumberType

_dbg = verbosity(0,name='node_exec');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

class editRequest(QDialog):
  def __init__(self, parent):
    self.parent = parent
    QDialog.__init__(self, parent._wtop, 'TEST', 1, 0)
    self.setCaption('Test')
#    self.setGeometry(30, 30, 400, 200)

#
# All input/text/button widgets are in rows -> enclosed in a VBox
#
    self.top = QVBoxLayout(self, 10, 5)

#
# All input/text widgets are in are in columns -> envlosed in an HBox
#
    self.UserIn = QHBoxLayout(self.top)

#
# Left side lables
#
    self.LeftLables = QVBoxLayout(self.UserIn, 5)
    lbl = QLabel(' ', self)
    self.LeftLables.addWidget(lbl)
    lbl = QLabel('Freq', self)
    self.LeftLables.addWidget(lbl)
    lbl = QLabel('Time', self)
    self.LeftLables.addWidget(lbl)

#
# Start values
#
    self.StartValues = QVBoxLayout(self.UserIn, 5)
    lbl = QLabel('start', self)
    self.StartValues.addWidget(lbl);
    self.X0 = QLineEdit(self)
    self.X0.setText(str(self.parent.f0))
    self.StartValues.addWidget(self.X0)
    self.Y0 = QLineEdit(self)
    self.Y0.setText(str(self.parent.t0))
    self.StartValues.addWidget(self.Y0)
    
#
# Number of steps
#
    self.StepValues = QVBoxLayout(self.UserIn, 5)
    lbl = QLabel('n-steps', self)
    self.StepValues.addWidget(lbl);
    self.Xn = QLineEdit(self)
    self.Xn.setText(str(self.parent.fn))
    self.StepValues.addWidget(self.Xn)
    self.Yn = QLineEdit(self)
    self.Yn.setText(str(self.parent.tn))
    self.StepValues.addWidget(self.Yn)

#
# Stop values
#
    self.StopValues = QVBoxLayout(self.UserIn, 5)
    lbl = QLabel('stop', self)
    self.StopValues.addWidget(lbl);
    self.X1 = QLineEdit(self)
    self.X1.setText(str(self.parent.f1))
    self.StopValues.addWidget(self.X1)
    self.Y1 = QLineEdit(self)
    self.Y1.setText(str(self.parent.t1))
    self.StopValues.addWidget(self.Y1)

#
# Variable Time
#
#    self.TimeRange = QHBoxLayout(self.top, 5)
#    lbl = QLabel('Time', self)
#    self.TimeRange.addWidget(lbl)
#    self.TimeEdit = QLineEdit(self)
#    self.TimeRange.addWidget(self.TimeEdit)

#
# Buttons
#
    self.Buttons = QHBoxLayout(self.top, 5)
    self.cmdOK = QPushButton('OK', self)
    QObject.connect(self.cmdOK, SIGNAL('clicked()'), self.slotcmdOK)
    self.Buttons.addWidget(self.cmdOK)
    self.cmdCancel = QPushButton('Cancel', self)
    QObject.connect(self.cmdCancel, SIGNAL('clicked()'), self.slotcmdCancel)
    self.Buttons.addWidget(self.cmdCancel)

    self.show()

  def slotcmdCancel(self):
    self.close();

  def slotcmdOK(self):
    f0 = float(str(self.X0.text()));
    f1 = float(str(self.X1.text()));
    fn = int(str(self.Xn.text()))

    if f0 == f1:
       f1 = f1 + 1;
    if fn == 0: fn = 1

    self.parent.f0 = f0;
    self.parent.f1 = f1;
    self.parent.fn = fn;

    t1 = float(str(self.Y1.text()))
    tn = int(str(self.Yn.text()))
    if tn == 0: tn = 1
    self.parent.tn = tn
#
# Allow three values for t0
#
    tvalue = str(self.Y0.text())
    s1 = split(tvalue)
    l = len(s1)
    t00 = float(s1[0])
    if l == 1:
      t01 = 0
      t02 = -1
    elif l == 2:
      t01 = 1
      t02 = float(s1[1])
    else:
      t01 = float(s1[1])
      t02 = float(s1[2])
    dt = t1 - t00
    if t02 == -1:
      self.parent.t0 = t00
      if t1 == t00:
        t1 = t1+1
      self.parent.t1 = t1
      self.parent.doNewRequest()
    else:
      t = t00
      while t <= t02:
        self.parent.t0 = t
        self.parent.t1 = self.parent.t0 + dt
        self.parent.doNewRequest()
        t = t + t01

#    self.close()

class startLoop(QDialog):
  def __init__(self, parent, name='LoopCnt', modal=0):
    self.parent = parent
    QDialog.__init__(self, parent._wtop, name, modal, 0)
    self.setCaption('LoopCnt')
    self.DTime = 1000;
    self.setGeometry(30, 30, 200, 100)

    self.v = QVBoxLayout(self, 10, 5)

    self.Xh = QHBoxLayout(self.v, 5)
    lbl = QLabel('X', self)
    self.Xh.addWidget(lbl)
    self.X0 = QLineEdit(self)
    self.X0.setText(str(self.DTime))
    self.Xh.addWidget(self.X0)

    self.buttons = QHBoxLayout(self.v, 5)
    self.cmdOK = QPushButton('OK', self)
    QObject.connect(self.cmdOK, SIGNAL('clicked()'), self.slotcmdOK)
    self.buttons.addWidget(self.cmdOK)

    self.cmdExit = QPushButton('Exit', self)
    QObject.connect(self.cmdExit, SIGNAL('clicked()'), self.slotExit)
    self.buttons.addWidget(self.cmdExit)

    self.doLoop = False;

    self.show()

  def slotcmdOK(self):
    if self.doLoop:
      self.killTimers();
      self.doLoop = False;
      self.cmdOK.setText('OK');
    else:
      x0 = int(str(self.X0.text()))
      self.DTime = x0;
      self.startTimer(self.DTime);
      self.cmdOK.setText('Stop');
      self.doLoop = True;

  def slotExit(self):
    self.killTimers();
    self.close();

  def timerEvent(self, event):
    self.parent._reexecute();

class NA_NodeExecute (NodeAction):
  text = "Reexecute";
  iconset = pixmaps.reexecute.iconset;
  def activate (self,node):
    Grid.Services.addDataItem(makeNodeDataItem(node,viewer=Executor));
    
  def is_enabled (self,node):
    # available in idle mode, or when stopped at a debugger. 
    # If debug_level is set, node must be idle (otherwise, we can't trust the
    # node control status to be up-to-date)
    return ( self.tb.is_stopped or not self.tb.is_running ) and \
           ( not self.tb.debug_level or node.is_idle() );

_request_type = dmi_type('MeqRequest',record);

class Executor (browsers.GriddedPlugin):
  viewer_name = "Executor";
  _icon = pixmaps.reexecute;

  def is_viewable (data):
    return isinstance(data,_request_type) or \
      isinstance(getattr(data,'request',None),_request_type);
  is_viewable = staticmethod(is_viewable);

  defaultOpenItems = ({'cells':({'grid':None,'domain':None},None)},None);

  def __init__(self,gw,dataitem,cellspec={},**opts):
    self.rid1 = 0
    self.rid2 = 0
    self.rid3 = 0
    self.rid4 = 0
    browsers.GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    _dprint(1,"started with",dataitem.udi,dataitem.data);
    self._has_data = False;
    self._request = None;
    self._wtop = QVBox(self.wparent());
  
    udi_root = dataitem.udi;
    if not udi_root.endswith('/request'): # not a direct request object
      udi_root += '/request';
    self.reqView = browsers.HierBrowser(self.wtop(),"","",udi_root=udi_root);
    QObject.connect(self.reqView.wtop(),PYSIGNAL("displayDataItem()"),
                    self.wtop(),PYSIGNAL("displayDataItem()"));
    self.reqView.set_refresh_func(dataitem.refresh_func);
    
    buttonbox = QHBox(self.wtop());

    self.buttonDef = QPushButton(buttonbox, "buttonDef")
    self.buttonDef.setText('&NewRequest')

    self.buttonLoop = QPushButton(buttonbox, "buttonLoop");
    self.buttonLoop.setText('&Loop');
    self.doLoop = False;

    self.buttonOk = QPushButton(buttonbox,"buttonOk");
    self.buttonOk.setText("&Execute");
    self.buttonOk.setIconSet(pixmaps.reexecute.iconset());
    self.buttonOk.setAccel(Qt.ALT+Qt.Key_E);
    
    buttonRefresh = QPushButton(buttonbox,"buttonRefresh");
    buttonRefresh.setText("&Refresh");
    buttonRefresh.setIconSet(pixmaps.refresh.iconset());
    buttonRefresh.setAccel(Qt.ALT+Qt.Key_R);
    
    QObject.connect(self.buttonDef,SIGNAL("clicked()"),self._newreq);
    QObject.connect(self.buttonLoop,SIGNAL("clicked()"),self._startLoop)
    QObject.connect(self.buttonOk,SIGNAL("clicked()"),self._reexecute);
    QObject.connect(buttonRefresh,SIGNAL("clicked()"),
                    self.wtop(),PYSIGNAL("refresh()"));
    QObject.connect(self.wtop(),PYSIGNAL("refresh()"),self._refresh);
    
    ### refresh initially disabled, until we get data in
    self.buttonOk.setEnabled(0);
    
    # form a caption and set contents
    (name,ni) = meqds.parse_node_udi(dataitem.udi);
    caption = "Reexecute <b>%s</b>" % (name or '#'+str(ni),);
    self._node = ni;
    self._name = name;
    # setup widgets
    self.set_widgets(self.wtop(),caption,icon=self.icon());
    
    # connect font signal 
    QObject.connect(self.wtop(),PYSIGNAL("fontChanged()"),
                      self.reqView.wtop(),PYSIGNAL("fontChanged()"));
    # add menu entry
    context_menu = self.cell_menu();
    if context_menu is not None:
      context_menu.insertSeparator();
      menu = browsers.PrecisionPopupMenu(context_menu,prec=self.reqView.get_precision());
      context_menu.insertItem(pixmaps.precplus.iconset(),'Number format',menu);
      QWidget.connect(menu,PYSIGNAL("setPrecision()"),\
                      self.reqView.set_precision);
    
    # set data
    if dataitem.data is not None:
      self.set_data(dataitem);

    
  def wtop(self):
    return self._wtop;

  def set_data (self,dataitem,**opts):
    _dprint(2,'set_data called, has_data is',self._has_data);
    if not self._has_data: 
      self._has_data = True;
      req = dataitem.data;
      # data is not a request, must be node state then
      if not isinstance(req,_request_type):
        req = getattr(req,'request',None);
        # if no request in node state, report this
        if not isinstance(req,_request_type): 
          if self._request is not None:
            QMessageBox.warning(self.wtop(),'Missing request',
                "No request field found in node state, keeping old request",
                QMessageBox.Ok,QMessageBox.NoButton);
          else:
            self.reqView.wlistview().setRootIsDecorated(False);
            self.reqView.clear();
            QListViewItem(self.reqView.wlistview(),'','','(no request found in node)');
          return;
      # ok, at this point req is a valid request object, load it
      self._request = req;
      self.reqView.wlistview().setRootIsDecorated(True);
      if not self._request.has_field('request_id'):
        self._request.request_id = hiid(0,0,0,0);
      self.rid1 = int(self._request.request_id[0]);
      self.rid2 = int(self._request.request_id[1]);
      self.rid3 = int(self._request.request_id[2]);
      self.rid4 = int(self._request.request_id[3]);
      self.buttonOk.setEnabled(True);
      self.reqView.set_content(self._request);
      self.reqView.set_open_items(self.defaultOpenItems);
      # enable cell and flash
      self.enable();
      self.flash_refresh();
      
  def _refresh (self):
    # clear the has_data flag, so that set_data above will accept new data
    _dprint(2,'refresh expected now');
    self._has_data = False;
    
  def _reexecute (self):
    if not self._request:
      return;
    _dprint(1,'accepted: ',self._request);
    self.rid1 = self.rid1 + 1
    self._request.request_id = hiid(self.rid1, self.rid2, self.rid3, self.rid4);
    
    if self._node:
      cmd = record(nodeindex=self._node,request=self._request,get_state=True);
    elif self._name:
      cmd = record(name=self._name,request=self._request,get_state=True);
    mqs().meq('Node.Execute',cmd,wait=False);
    self._refresh();

  def _startLoop(self):
    startLoop(self)

  def _newreq(self):
    if self._request == None:
      self.f0 = 0
      self.f1 = 1
      self.fn = 10
      self.t0 = 0
      self.t1 = 1
      self.tn = 10
    else:
      self.f0 = self._request.cells.domain.freq[0];
      self.f1 = self._request.cells.domain.freq[1];
      if self._request.cells.segments.freq.start_index == self._request.cells.segments.freq.end_index:
        self.fn = 1
      else:
        self.fn = len(self._request.cells.cell_size.freq);
      self.t0 = self._request.cells.domain.time[0];
      self.t1 = self._request.cells.domain.time[1];
      if self._request.cells.segments.time.start_index == self._request.cells.segments.time.end_index:
        self.tn = 1
      else:
        self.tn = len(self._request.cells.cell_size.time);
    editRequest(self);
    self._refresh();

  def doNewRequest(self):
    newd = meq.domain(self.f0, self.f1, self.t0, self.t1);
    newc = meq.cells(domain=newd, num_freq=self.fn, num_time=self.tn);
    self._request = meq.request(cells=newc);
    self.rid1 = self.rid1 + 1
    self.rid2 = self.rid2 + 1
    self.rid3 = self.rid3 + 1
    self.rid4 = self.rid4 + 1
    self._request.request_id = hiid(self.rid1, self.rid2, self.rid3, self.rid4);
    if self._node:
      cmd = record(nodeindex=self._node,request=self._request,get_state=True);
    elif self._name:
      cmd = record(name=self._name,request=self._request,get_state=True);
    mqs().meq('Node.Execute',cmd,wait=False);
    self._refresh();


def define_treebrowser_actions (tb):
  _dprint(1,'defining node-execute treebrowser actions');
  tb.add_action(NA_NodeExecute,30,where="node");

Grid.Services.registerViewer(_request_type,Executor,priority=25);
Grid.Services.registerViewer(meqds.NodeClass(),Executor,priority=25);




from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

from MeqGUI.GUI import browsers
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI import Grid

try:
  from .TDLEditor_qscintilla import TDLEditor
except:
  from .TDLEditor_standard import TDLEditor
  
from .TDLErrorFloat import TDLErrorFloat
from .TDLOptionsDialog import TDLOptionsDialog



class TDLFileDataItem (Grid.DataItem):
  """represents a GridDataItem for a TDL script""";
  def __init__ (self,pathname):
    # read the file (exception propagated outwards on error)
    ff = open(pathname);
    text = ff.read();
    ff.close();
    basename = os.path.basename(pathname);
    # create the item
    udi = '/tdlfile/'+pathname;
    name = basename;
    caption = '<b>'+basename+'</b>';
    desc = 'TDL file '+pathname;
    Grid.DataItem.__init__(self,udi,name=name,caption=caption,desc=desc,data=text,viewer=TDLBrowser,refresh=None);
    # add extra pathname attribute for tdl objects
    self.tdl_pathname = pathname;


class TDLBrowser(browsers.GriddedPlugin):
  _icon = pixmaps.text_tdl;
  viewer_name = "TDL Browser";

  def __init__(self,gw,dataitem,cellspec={},default_open=None,**opts):
    browsers.GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    self._wedit = TDLEditor(self.wparent());
    self.set_widgets(self.wtop(),dataitem.caption,icon=self.icon());
    if dataitem.data is not None:
      self.set_data(dataitem);
    QObject.connect(self.wtop(),PYSIGNAL("fontChanged()"),self.wtop().adjust_editor_font);

  def wtop (self):
    return self._wedit;
  def editor (self):
    return self._wedit;

  def set_data (self,dataitem,default_open=None,**opts):
    _dprint(3,'set_data ',dataitem.udi);
    pathname = getattr(dataitem,'tdl_pathname',None);
    self._wedit.load_file(pathname,text=dataitem.data);

  def highlight (self,color=True):
    browsers.GriddedPlugin.highlight(self,color);
    self._wedit.has_focus(bool(color));

Grid.Services.registerViewer(str,TDLBrowser,priority=10,check_udi=lambda x:x.endswith('_tdl'));




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from MeqGUI.GUI.app_proxy_gui import *
from MeqGUI.GUI.pixmaps import pixmaps
from Timba.Meq import meqds
from Timba.Meq.meqds import mqs
from MeqGUI.GUI import meqgui

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL,ClickableTreeWidget,BusyIndicator
from MeqGUI.GUI.widgets import DataDraggableTreeWidget
import Kittens

import weakref
import re
import copy
import math

_dbg = verbosity(0,name='tb');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;


class AppState (object):
  Idle       = -hiid('idle').get(0);
  Stream     = -hiid('stream').get(0);
  Debug      = -hiid('debug').get(0);
  Execute    = -hiid('execute').get(0);
  
class StickyTreeWidgetItem (QTreeWidgetItem):
  """This is a QTreeWidgetItem that ignores sorting, 'sticking' to one
  place in the listview.
  Each such item is created with a 'key' argument that determines its 
  placement in the listview, regardless of sort column or order. Items 
  with keys>0 are placed after regular items, items with keys<0 are 
  placed before regular items.
  """;
  def __init__ (self,parent,name=None,**kw):
    self._key = kw.get('key',1);
    QTreeWidgetItem.__init__(self,parent);
    name and self.setText(0,name);
    self.setFirstColumnSpanned(True);
  def __lt__ (self,other):
    try: return self._key < other._key;
    except AttributeError:  # other item not keyed
      return self._key < 0;
  def __ge__ (self,other):
    return other < self;

class TreeBrowser (QObject):

  class NodeItem (QTreeWidgetItem):
    def __init__(self,tb,node,name,parent=None,after=None,stepchild=False):
      QTreeWidgetItem.__init__(self,parent,after)
      self.setText(0,name);
      _dprint(3,"creating TreeBrowser.NodeItem",name);
      # fill basic listview stuff
      self.setText(tb.icolumn("class"),str(node.classname));
      self.setText(tb.icolumn("index"),str(node.nodeindex));
      self.setTextAlignment(tb.icolumn("index"),Qt.AlignRight);
      if tb._mpi_num_proc > 1:
        self.setText(tb.icolumn("proc"),"P%d"%node.proc);
      self.setFlags(Qt.ItemIsSelectable|Qt.ItemIsEnabled|Qt.ItemIsDragEnabled);
      if node.children or node.step_children:
        self.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);
      # external state
      self.tb = weakref_proxy(tb);      # self.tb returns treebrowser
      self.node = weakref_proxy(node);  # self.node returns meqds node
      # internal state, etc.
      self._expanded = False;
      self._stopped = 0;
      self._udi  = meqds.node_udi(node);
      self._callbacks = [];
      self._menu_actions = [];
      self._stepchild = stepchild;
      self._visual_status = None;
      self._default_background = self.background(0);
      self._default_foreground = self.foreground(0);
      self._column_icons = {};
      # add ourselves to the item list for this node
      if not hasattr(node,'_tb_items'):
        node._tb_items = [];
      node._tb_items.append(weakref.ref(self));
      # subsribe to various events
      node.subscribe_status(self._update_status);
      node.subscribe_state(self._update_state);
      QObject.connect(node,PYSIGNAL("update_debug()"),self._update_debug);
      # default color group is None to use normal colors
      # make sure pixmaps, etc. are updated
      self._update_status(node,node.control_status);
      # add click handlers
      self._item_event_handler = {};
      # left-click: view
      self._item_event_handler[('click',1)] = xcurry(self._add_node_viewer);
      # middle-click: new view
      self._item_event_handler[('click',4)] = xcurry(self._add_node_viewer,newcell=True);
      self._item_event_handler['menu'] = self._show_context_menu;
      
    def __del__ (self):
      self.cleanup();
    
    def setIcon (self,column,icon):
      # sets icon in column; checks if the previous setting is the same
      oldicon = self._column_icons.get(column);
      if oldicon is not icon:
        QTreeWidgetItem.setIcon(self,column,icon);
        self._column_icons[column] = icon;
    
    def cleanup (self):
      self._callbacks = self._item_event_handler = None;
      # try: 
      #  node = self.node;
      #  node._tb_items = None; # NB: kludge
      #  node.unsubscribe_status(self._update_status);
      #  node.unsubscribe_state(self._update_state);
      
    def _show_context_menu (self,point,col):
      if col in (self.tb.icolumn("execstate"),self.tb.icolumn("breakpoint")):
        menu = self.debug_menu();
      else:
        menu = self.context_menu();
      if menu is not None:
        menu.exec_(self.treeWidget().mapToGlobal(point));
      
    def _add_node_viewer (self,viewer=None,*dum,**kws):
      Grid.addDataItem(meqgui.makeNodeDataItem(self.node,viewer),show_gw=True,**kws);
      
    def _update_state (self,node,state,event):
      """updates node item based on changes in state.""";
      # flash the publish pixmap briefly to indicate new state
      # self._publish_pixmap is the "normal" (None or publish) pixmap set by _update_status()
      # below; we set a timer event to revert to it
      self.setIcon(self.tb.icolumn("publish"),pixmaps.publish_active.icon());
      QTimer.singleShot(500,self.xcurry(self.setIcon,(self.tb.icolumn("publish"),self._publish_icon)));
              
    def _update_debug (self,node):
      """updates node color based on its debugging status.""";
      background,foreground = self.tb.get_qbrushes(self._visual_status,self._stopped);
      self.setBackground(0,background or self._default_background);
      self.setForeground(0,foreground or self._default_foreground);
      
    def _update_status (self,node,old_status):
      """updates node item based on node status.""";
      tb = self.tb;
      control_status = node.control_status;
      # choose a colorgroup name for the item based on its status
      stopped = bool(control_status&meqds.CS_STOP_BREAKPOINT);
      result_status = control_status&meqds.CS_RES_MASK;
      if result_status == meqds.CS_RES_FAIL:
        self._visual_status = "fail";
      elif not control_status&meqds.CS_ACTIVE or \
           result_status in (meqds.CS_RES_WAIT,meqds.CS_RES_MISSING): 
        self._visual_status = "disabled";
      elif self._stepchild:
        self._visual_status = "stepchild";
      else:
        self._visual_status = None;
      ## # update status column
      ## self.setText(tb.icolumn("status"),node.control_status_string);
      # update breakpoint status pixmaps
      if control_status&meqds.CS_BREAKPOINT:
        self.setIcon(tb.icolumn("breakpoint"),pixmaps.breakpoint.icon());
      elif control_status&meqds.CS_BREAKPOINT_SS:
        self.setIcon(tb.icolumn("breakpoint"),pixmaps.forward_to.icon());
      else:
        self.setIcon(tb.icolumn("breakpoint"),QIcon());
      # update first column pixmaps for stopped and stopped-at-breakpoint nodes
      if control_status&meqds.CS_STOPPED:
        if control_status&meqds.CS_STOP_BREAKPOINT:
          self._stopped = 2;
          #self.setIcon(0,pixmaps.breakpoint.icon());
        else:
          self._stopped = 1;
        self.setIcon(0,pixmaps.clear_red_right.icon());
      else:
        self._stopped = 0;
        self.setIcon(0,QIcon());
      # call update_debug to complete setting of color group (debug state
      # determines background)
      self._update_debug(node);
      # update exec state pixmaps and rqid string
      es = meqds.CS_ES_state(control_status);
      self.setIcon(tb.icolumn("execstate"),es[4].icon());
      if node.request_id is None:
        self.setText(tb.icolumn("execstate"),'');
      else:
        self.setText(tb.icolumn("execstate"),str(node.request_id));
      # update result status pixmap
      icol = tb.icolumn("result");
      if result_status == meqds.CS_RES_NONE:
        self.setIcon(icol,QIcon());
      elif result_status == meqds.CS_RES_OK:
        if control_status&meqds.CS_RETCACHE:
          if control_status&meqds.CS_CACHED:
            self.setIcon(icol,pixmaps.blue_round_reload_return.icon());
          else:
            self.setIcon(icol,pixmaps.green_return.icon());
        else:
          if control_status&meqds.CS_CACHED:
            self.setIcon(icol,pixmaps.blue_round_reload.icon());
          else:
            self.setIcon(icol,pixmaps.green_return.icon());
      elif result_status == meqds.CS_RES_WAIT:
        self.setIcon(icol,pixmaps.blue_round_clock.icon());
      elif result_status == meqds.CS_RES_EMPTY:
        if control_status&meqds.CS_RETCACHE:
          self.setIcon(icol,pixmaps.blue_round_empty_return.icon());
        else:
          self.setIcon(icol,pixmaps.blue_round_empty.icon());
      elif result_status == meqds.CS_RES_MISSING:
        self.setIcon(icol,pixmaps.grey_round_cross.icon());
      elif result_status == meqds.CS_RES_FAIL:
        self.setIcon(icol,pixmaps.red_round_cross.icon());
      # update enabled/disabled pixmap
      if control_status&meqds.CS_ACTIVE:
        self.setIcon(tb._icol_disable,QIcon());
      else:
        self.setIcon(tb._icol_disable,pixmaps.cancel.icon());
      # update publish/state pixmap
      if control_status&meqds.CS_PUBLISHING:
        self._publish_icon = pixmaps.publish.icon();
      else:
        self._publish_icon = QIcon();
      self.setIcon(tb.icolumn("publish"),self._publish_icon);
      # update breakpoints menu
      try: bp_actions = self._debug_bp_actions;
      except AttributeError: pass;
      else:
        _dprint(3,'node',self.node.name,'breakpoint mask is',self.node.breakpoint);
        self._set_breakpoints_quietly = True;
        for (qa,bp) in bp_actions:
          qa.setChecked((self.node.breakpoint&bp)!=0);
        self._set_breakpoints_quietly = False;
    
    def expand (self):
      if self._expanded:
        return;
      i1 = self;
      busy = BusyIndicator();
      items = [];
      # sort children into proper order
      children = list(self.node.children);
      from functools import cmp_to_key
      children.sort(key=cmp_to_key(self._compare_children));
      # generate items for each child
      for (key,ni) in children:
        if ni > 0:
          node = meqds.nodelist[ni];
          name = self.node.child_label_format() % (key,node.name);
          i1 = self.__class__(self.tb,node,name);
        else:  # missing child, generate dummy item with <none> in it
          name = self.node.child_label_format() % (key,'<none>');
          i1 = QTreeWidgetItem();
          i1.setText(0,name);
        setattr(i1,'_sort_label',key);
        items.append(i1);
      for ni in self.node.step_children:
        node = meqds.nodelist[ni];
        name = "(" + node.name +")";
        items.append(self.__class__(self.tb,node,name,stepchild=True));
      self.addChildren(items);
      self._expanded = True;
      
    @staticmethod
    def _compare_children (a,b):
      a,b = a[0],b[0];
      # try to convert to integer
      try:  a = int(a);
      except ValueError: pass;
      try:  b = int(b);
      except ValueError: pass;
      from past.builtins import cmp
      if isinstance(a,int):
	      if isinstance(b,int):
	        return cmp(a,b);
	      else:
	        return 1;
      elif isinstance(b,int):
	      return -1;
      else:
	      return cmp(a,b);
	
    def curry (self,*args,**kwargs):
      cb = curry(*args,**kwargs);
      self._callbacks.append(cb);
      return cb;
      
    def xcurry (self,*args,**kwargs):
      cb = xcurry(*args,**kwargs);
      self._callbacks.append(cb);
      return cb;
      
    def _fill_menu (self,menu,which,separator=False):
      """adds node-related actions defined in the TreeBrowser to the
      specified menu."""
      acts = self.tb.get_action_list(which);
      acts.sort(); # sorts by priority
      for (pri,action) in acts:
        _dprint(5,'action',action,'of type',type(action).__name__);
        if action is None:                # add separator
          separator = True;
        elif isinstance(action,QAction):  # add regular action 
          if separator:
            menu.addSeparator();
            separator = False;
          menu.addAction(action);
        elif issubclass(action,NodeAction):            # assume NodeAction
          if action.applies_to_node(self.node):
            # instantiate action, add to menu if successful
            try:
              na = action(self,menu,separator=separator);
            except:
              _dprint(1,'error adding action to node menu:',sys.exc_info());
            else:
              self._menu_actions.append(na);
              separator = False;
        else:
          raise TypeError('unknown action type '+type(action).__name__);
    
    def _set_node_breakpoint (self,node,mask,set=True):
      if self._set_breakpoints_quietly:
        return;
      if set:
        meqds.set_node_breakpoint(node,mask);
      else:
        meqds.clear_node_breakpoint(node,mask);
    
    def debug_menu (self):
      try: menu = self._debug_menu;
      except AttributeError:
        self._set_breakpoints_quietly = False;
        node = self.node;
        menu = self._debug_menu = QMenu("Debugging",self.treeWidget());
        menu.setIcon(pixmaps.breakpoint.icon());
        # menu1 = self._debug_bp_menu = QPopupMenu();
        self._debug_bp_actions = [];
        _dprint(3,'node',node.name,'breakpoint mask is',node.breakpoint);
        #label = QHBox(menu);
        #label1 = QLabel(label);
        #label1.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum);
        #label1.setPixmap(pixmaps.breakpoint.pm());
        #label2 = QLabel("Set breakpoint at:",label);
        #label2.setAlignment(Qt.AlignLeft);
        #label2.setIndent(10);
        #label2.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Minimum);
        #menu.insertItem(label);
        menu.addAction(pixmaps.breakpoint.icon(),"Set breakpoint at:");
	      # Kittens.widgets.addMenuLabel(menu,"Set breakpoint at:");
        for st in meqds.CS_ES_statelist:
          title = ''.join(('    ',node.name,':',st[1]));
          bpmask = meqds.breakpoint_mask(st[0]);
          qa = menu.addAction(st[4].icon(),title,self.curry(self._set_node_breakpoint,node.nodeindex,bpmask));
          qa.setCheckable(True);
          qa.setChecked((node.breakpoint&bpmask)!=0);
          self._debug_bp_actions.append((qa,bpmask));
        qa = menu.addAction(pixmaps.red_round_cross.icon(),''.join(('    ',node.name,':FAIL')), 
                                                                    self.curry(self._set_node_breakpoint,node.nodeindex,meqds.BP_FAIL));
        qa.setCheckable(True);
        qa.setChecked((node.breakpoint&meqds.BP_FAIL)!=0);
        self._debug_bp_actions.append((qa,bpmask));
        menu.addAction(pixmaps.node_any.icon(),'    all of the above', self.curry(self._set_node_breakpoint,node.nodeindex,meqds.BP_ALL));
        # menu.insertItem(pixmaps.breakpoint.icon(),"Set &breakpoint at",menu1);
        menu.addSeparator();
        menu.addAction(pixmaps.roadsign_nolimit.icon(),"Clear &all breakpoints at "+node.name,self.xcurry(\
              meqds.clear_node_breakpoint,(node.nodeindex,meqds.BP_ALL),_argslice=slice(0)));
        self._fill_menu(menu,"debug",separator=True);
      return menu;
      
    def context_menu (self):
      try: menu = self._context_menu;
      except AttributeError:
        # create menu on the fly when first called for this item
        node = self.node;
        menu = self._context_menu = QMenu(self.treeWidget());
        # insert title
        Kittens.widgets.addMenuLabel(menu,"""<b>%s</b> 											<i><small>(%s)</small></i>"""%(node.name,node.classname));
        # menu.insertItem(pixmaps.info_blue_round.icon(),"Show icon reference...",self.tb.show_icon_reference);
        # insert viewer list submenus
        viewer_list = Grid.Services.getViewerList(meqds.NodeClass(node.classname));
        if viewer_list: 
          # create display submenus
          menu1 = self._display_menu1 = menu.addMenu(pixmaps.viewmag.icon(),"Display with");
          menu2 = self._display_menu2 = menu.addMenu(pixmaps.viewmag_plus.icon(),"New display with");
          for v in viewer_list:
            # create entry for viewer
            name = getattr(v,'viewer_name',v.__name__);
            try: icon = v.icon();
            except AttributeError: icon = QIcon();
            # add entry to both menus ("Display with" and "New display with")
            menu1.addAction(icon,name,self.xcurry(self._add_node_viewer,viewer=v));
            menu2.addAction(icon,name,self.xcurry(self._add_node_viewer,viewer=v,newcell=True));
        # add node actions
        self._fill_menu(menu,"node",separator=True);
        # add debugging menu
        menu.addMenu(self.debug_menu());
      # refresh all node actions
      list(map(lambda a:a.update(menu,self.node),self._menu_actions));
      return menu;
      
    def paintCell (self,painter,cg,column,width,align):
      return QTreeWidgetItem.paintCell(self,painter,self._color_group or cg,column,width,align);

    def clear_children (parent):
      """Recursively cleans up all child items of the given parent, in
      preparation for deleting them. This is mainly done to remove cyclic
      references caused by the NodeItem._callbacks list.""";
      if isinstance(parent,QTreeWidget):
        children = [ parent.topLevelItem(i) for i in range(parent.topLevelItemCount()) ];
      else:
        children = [ parent.child(i) for i in range(parent.childCount()) ];
      for item in children:
        _dprint(2,item.text(0),item);
        try: item.cleanup();
        except AttributeError: pass;
        TreeBrowser.NodeItem.clear_children(item);
    clear_children = staticmethod(clear_children);
    
  ######################################################################################################
  # start of TreeBrowser implementation    
  ######################################################################################################
  def __init__ (self,parent):
    QObject.__init__(self);
    self._parent = weakref.proxy(parent);
    # ---------------------- construct GUI
    self._tw = self._wtop = tw = DataDraggableTreeWidget(ClickableTreeWidget)(parent);
    tw.setSortingEnabled(False);
    tw.header().setSortIndicatorShown(False);
    tw.setRootIsDecorated(True);
    tw.setIndentation(12);
    # tw.setSorting(-1);
    tw.header().setResizeMode(QHeaderView.Interactive);
    tw.setFocus();
    QObject.connect(tw,SIGNAL('itemExpanded(QTreeWidgetItem*)'),self._expand_node);
    QObject.connect(tw,PYSIGNAL('mouseButtonClicked()'),self._item_clicked);
    QObject.connect(tw,PYSIGNAL('itemContextMenuRequested()'),self._show_context_menu);
    QObject.connect(tw,SIGNAL('currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)'),self._set_recent_item);
    # setup columns
    tw.header().setDefaultAlignment(Qt.AlignRight);
    tw.setColumnCount(8);
    self._tw_headeritem = QTreeWidgetItem();
    self._column_map = {};
    self.add_column("node");
    self.add_column("publish",'',24,icon=pixmaps.publish.icon());
    self.add_column("breakpoint",'',24,icon=pixmaps.breakpoint.icon());
    self.add_column("execstate","xs/rqid");
    self.add_column("result",'',24,icon=pixmaps.blue_round_result.icon());
    self.add_column("proc",'',24);
    self._icol_disable = self.add_column("class");
    self.add_column("index",width=60);
    self._tw.setHeaderItem(self._tw_headeritem);
    self._tw.header().hide();
    
    # map the get_drag_item methods
    tw.get_drag_item = self.get_drag_item;
    tw.get_drag_item_type = self.get_drag_item_type;
    #---------------------- setup QBrushes for nodes
    self._qbrushes = {};
    # brushes used to indicate state of nodes
    status_brushes = {  None:         None,
                        'stepchild':  QBrush(QColor("darkgrey")),
                        'disabled':   QBrush(QColor("lightgray")),
                        'fail':       QBrush(QColor("red")) };
    stopped_brushes = { 0:  None,
                        1:  QBrush(QColor("lightblue")), 
                        2:  QBrush(QColor("skyblue")) };
    for status,qb1 in status_brushes.items():
      for stopped,qb2 in stopped_brushes.items():
        self._qbrushes[status,stopped] = qb2,qb1; 
    # ---------------------- setup toolbars, QActions, menus, etc.
    # scan all modules for define_treebrowser_actions method, and call them all
    self._actions = {};
    funcs = set();
    for (name,mod) in list(sys.modules.items()):
      _dprint(4,'looking for treebrowser actions in',name);
      try: 
        if (name.startswith("Timba") or name.startswith("MeqGUI")) and \
          callable(mod.define_treebrowser_actions):
          _dprint(3,'tb action found in',name,'adding to set');
          funcs.add(mod.define_treebrowser_actions);
      except AttributeError: pass;
    _dprint(1,len(funcs),'unique treebrowser action-definition methods found');
    for f in funcs:
      f(self);
    # --- now, build up toolbar from defined actions
    mainwin = parent;
    while mainwin and not isinstance(mainwin,QMainWindow):
      mainwin = mainwin.parent();
    self._toolbar = QToolBar(mainwin);
    self._toolbar.setIconSize(QSize(16,16));
    self._toolbar_actions = [];
    tba = self.get_action_list("toolbar");
    import six
    if six.PY3:
      tba.sort(key=lambda x: x[0])
    else:
      tba.sort();
    prev_sep = True; # flag: previous entry is a separator, true at top
    for (pri,action) in tba:
      if action is None:
        if not have_sep: 
          self._toolbar.addSeparator();
          have_sep = True;
      elif action == "stretch":
        stretch = QWidget(self._toolbar);
        # self._toolbar.setStretchableWidget(stretch);
      elif isinstance(action,QAction):
        self._toolbar_actions.append(action);
        self._toolbar.addAction(action);
        action.setEnabled(False);
        have_sep = False;
    # ---------------------- connect to meqds forest list
    QObject.connect(meqds.nodelist,PYSIGNAL("cleared()"),self.clear);
    QObject.connect(meqds.nodelist,PYSIGNAL("loaded()"),self.update_nodelist);
    QObject.connect(meqds.nodelist,PYSIGNAL("requested()"),self._requested_nodelist);

    # ---------------------- other internal state
    self._forest_breakpoint = 0;
    self._recent_item = None;
    self._callbacks = [];
    self._stopped_nodes = [];
    self._breakpoint_nodes = [];
    self._mpi_num_proc = 1;
    #----------------------- public state
    self.app_state = None;
    self.debug_level = 0;
    self.is_connected = self.is_loaded = self.is_running = self.is_stopped = False;
    
  def update_nodelist (self):
    # clear view
    _dprint(1,"updating node list");
    self.clear();
    self.is_loaded = True;
    self._update_all_controls();
    # add forest state and bookmarks
    self._fst_item = StickyTreeWidgetItem(self._tw,name="Forest state",key=5);
    self._fst_item.setIcon(0,pixmaps.view_tree.icon());
    self._fst_item._udi = '/forest';
    self._fst_item.setFlags(Qt.ItemIsSelectable|Qt.ItemIsEnabled|Qt.ItemIsDragEnabled);
    self._fst_item._item_event_handler = {};
   # left-click: view
    self._fst_item._item_event_handler[('click',1)] = xcurry(self._view_forest_state);
    # middle-click: new view
    self._fst_item._item_event_handler[('click',4)] = xcurry(self._view_forest_state,newcell=True);
    # self._fst_item.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);
    
    # self._bkmark_item = StickyTreeWidgetItem(self._tw,"Bookmarks",key=60);
    # self._bkmark_item.setPixmap(0,pixmaps.bookmark.pm());
    # add nodelist views
    nodelist = meqds.nodelist;
    self._recent_item = None;
    # add 'All Nodes' item
    #all_item  = StickyTreeWidgetItem(self._tw,"All nodes (%d)"%len(nodelist),key=10);
    #all_item._no_auto_open = True;
    #all_item._iter_nodes = nodelist.iternodes();
    #all_item.setFlags(Qt.ItemIsEnabled);
    # add 'Root Nodes' item
    from past.builtins import cmp
    from functools import cmp_to_key
    rootnodes = sorted(nodelist.rootnodes(),key=cmp_to_key(lambda a,b:cmp(a.name,b.name)));
    rootnodes.sort(key=lambda x: x.objectName());
    rootitem  = self._tw_rootitem = \
      StickyTreeWidgetItem(self._tw,name="Root nodes (%d)"%len(rootnodes),key=30);
    rootitem._iter_nodes = iter(rootnodes);
    rootitem.setFlags(Qt.ItemIsEnabled);
    rootitem.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);
    # add 'By Class' item
    classes = nodelist.classes();
    cls_item  = item = \
      StickyTreeWidgetItem(self._tw,name="By class (%d)"%len(classes),key=20);
    cls_item._no_auto_open = True;
    cls_item.setFlags(Qt.ItemIsEnabled);
    items = [];
    for (cls,nodes) in sorted(iter(classes.items()),key=cmp_to_key(lambda a,b:cmp(a[0],b[0]))):
      item = QTreeWidgetItem();
      item.setText(0,"%s (%d)"%(cls,len(nodes)));
      item.setText(self.icolumn("class"),cls);
      item.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);
      nodes = sorted(nodes,key=cmp_to_key(lambda a,b:cmp(a.name,b.name)));
      item._iter_nodes = iter(nodes);
      item._no_auto_open = True;
      items.append(item);
    cls_item.addChildren(items);
    # add 'By Processor' item
    if self._mpi_num_proc > 1:
      procitem = item = \
        StickyTreeWidgetItem(self._tw,"By processor (%d)"%self._mpi_num_proc,key=25);
      procitem._no_auto_open = True;
      # add node to per-processor list only if it is a root node of
      # that processor (i.e. none of its parents belong to that processor)
      for proc in range(self._mpi_num_proc):
        # list of all nodes on processor
        proclist = [x for x in nodelist.iternodes() if x.proc == proc]; 
        # list of root nodes of that processor
        procrootlist = sorted([x for x in proclist if not [y for y in x.parents if nodelist[y].proc==proc]],
                              key=cmp_to_key(lambda a,b:cmp(a.name,b.name)));
        item = QTreeWidgetItem(procitem,item);
        item.setText(0,"P%d (%d)"%(proc,len(proclist)));
        item.setFirstColumnSpanned(True);
        item.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);
        item._iter_nodes = iter(procrootlist);
        item._no_auto_open = True;
    # resort
    self._tw.sortItems(0,Qt.AscendingOrder);
    # emit signal
    self.emit(SIGNAL("forestLoaded"));
      
  def _requested_nodelist (self):
    _dprint(2,"nodelist requested");
    if not self.is_loaded:
      self._tw.clear();
      item = QTreeWidgetItem(self._tw);
      item.setText(0,"updating, please wait...");
     
  def clear (self):
    _dprint(2,"clearing tree browser");
    self.is_loaded = False;
    self._tw_rootitem = self._fst_item = self._bkmark_item = \
      self._recent_item = None;
    self.NodeItem.clear_children(self._tw);
    self._tw.clear();
    self._update_all_controls();
    
    
  def xcurry (self,*args,**kwargs):
    cb = xcurry(*args,**kwargs);
    self._callbacks.append(cb);
    return cb;
    
  def add_column (self,colname,caption=None,width=-1,icon=None):
    """adds column to tree widget. Colname is column name, caption is displayed
    name (same as colname by default), width is >=0 for fixed width""";
    icol = len(self._column_map);
    if caption is None:
      caption = colname;
    icon and self._tw_headeritem.setIcon(icol,icon);
    caption and self._tw_headeritem.setText(icol,caption);
    # self._tw.header().setMinimumSectionSize(icol,width);
    self._column_map[colname] = (icol,width);
    self._tw.header().setResizeMode(icol,QHeaderView.ResizeToContents);
    return icol;
    
  def icolumn (self,colname):
    """returns column number for a given colname""";
    return self._column_map[colname][0];
    
  def show_column (self,colname,show=True):
    """shows or hides the column specified by name""";
    try: (icol,width) = self._column_map[colname];
    except KeyError as AttributeError: return;
    # make column visible or not
    if show: 
      self._tw.header().showSection(icol);
    else:
      self._tw.header().hideSection(icol);
    #if width >= 0:
        #self._tw.setColumnWidth(icol,width);
      #else:
        #self._tw.setColumnWidthMode(icol,QTreeWidget.Maximum);
    #else:
      #self._tw.setColumnWidthMode(icol,QTreeWidget.Manual);
      #self._tw.setColumnWidth(icol,0);
      
  def _update_all_controls (self):
    """updates state of toolbar and other controls based on app state""";
    _dprint(3,self.debug_level,self.is_connected,self.is_running,self.is_stopped);
    # enable/disable toolbar buttons according to their methods
    for act in self._toolbar_actions:
      # call _is_visible() to setup visibility
      try:
        visible = act._is_visible();
      except AttributeError:  pass;
      except:
        _dprint(0,'exception while calling _is_visible on tb action:',sys.exc_info());
        traceback.print_exc();
      else:
        _dprint(4,'action',act.text(),'visible:',visible);
        act.setVisible(visible);
      # call _is_enabled() to setup enabledness
      try:
        enable = act._is_enabled();
        _dprint(4, act.text(), "enable",enable);
      except AttributeError: 
        _dprint(4, act.text(), "no _is_enabled() method");
        pass;
      except:
        _dprint(0,'exception while calling _is_enabled on tb action:',sys.exc_info());
        traceback.print_exc();
      else:
        _dprint(4,'action',act.text(),'enabled:',enable);
        act.setEnabled(enable);
    # set the verbosity levels
    self._setting_debug_control = True;
    lvl = max(0,min(self.debug_level,len(self._qa_dbg_verbosities)-1));
    self._qa_dbg_verbosities[lvl].setChecked(True);
    self._setting_debug_control = False;
    # set fail-breakpoint on or of
    self._qa_bp_on_fail.setChecked(self._forest_breakpoint&meqds.BP_FAIL);
    self._setting_debug_control = False;
    # show/hide the breakpoint column
    self.show_column("breakpoint");
    self.show_column("result");
    # show/hide the execstate column
    self.show_column("execstate",self.is_running);

  def get_qbrushes (self,status,stopped=0):
    return self._qbrushes[status,int(stopped)];
    
  def get_drag_item (self,key):
    if key == '/forest':
      return meqgui.makeForestDataItem();
    elif key.startswith('/node'):
      (name,ni) = meqds.parse_node_udi(key);
      node = meqds.nodelist[name or ni];
      return meqgui.makeNodeDataItem(node);
      
  def get_drag_item_type (self,key):
    return MeqGUI.Grid.DataItem;
 
  def wtop (self):
    return self._wtop;
    
  def wtoolbar (self):
    return self._toolbar;
    
  def connected (self,conn,auto_request=True):
    self.emit(SIGNAL("connected"),conn,);
    self.is_connected = conn;
    if conn is True:
      self._update_all_controls();
      if auto_request:
        self._request_nodelist();
    else:
      self.clear();

  def update_app_state (self,state):
    self.app_state = state;
    # self.is_running = state in (AppState.Debug,AppState.Stream,AppState.Execute);
    # self.is_stopped = state == AppState.Debug;
    # self._update_all_controls();
    
  def update_forest_status (self,fst):
    """Updates forest status: enables/disables debug QActions as appropriate,
    sets/clears highlighting of stopped nodes."""
    _dprint(2,"updating forest status");
    self.is_running = fst.executing;
    was_stopped = self.is_stopped;
    self.is_stopped = fst.stopped;
    self.debug_level = fst.debug_level;
    self._mpi_num_proc = getattr(fst,'mpi_num_proc',1);
    self.show_column("proc",self._mpi_num_proc>1);
    self._forest_breakpoint = fst.breakpoint;
    self.wtop().emit(SIGNAL("isRunning"),self.is_running,);
    self.wtop().emit(SIGNAL("isStopped"),self.is_stopped,);
    # if just stopped in the tree debugger, make sure all stopped nodes
    # are visible
    _dprint(5,"was stopped",was_stopped,"is stopped",self.is_stopped);
    if not was_stopped and self.is_stopped:
      # figure out list of stopped nodes
      self._stopped_nodes = [node for node in meqds.nodelist.iternodes() if node.is_stopped()];
      _dprint(5,len(self._stopped_nodes),"nodes stopped");
      self._breakpoint_nodes = [];
      for node in self._stopped_nodes:
        self.make_node_visible(node);
        if node.control_status&meqds.CS_STOP_BREAKPOINT:
          self._breakpoint_nodes.append(node);
      # nodes actually stopped at a breakpoint get priority in visibility
      for node in self._breakpoint_nodes:
        self.make_node_visible(node);
    # reset all controls
    self._update_all_controls();
        
  def is_node_visible (self,node):
    for itemref in getattr(node,'_tb_items',[]):
      if itemref() and itemref().isVisible():
          return True;
    return False;
    
  def make_node_visible (self,node):
    # no node list -- do nothing
    if not meqds.nodelist:
      return;
    # Go over every available item and find one with the lowest "display cost".
    # Cost starts at 0 and is incremented for every level that needs to be opened
    # to make the item visible. Thus, already-visible items will have a cost of 0.
    (best_cost,best_item) = (100000,None);
    for itemref in getattr(node,'_tb_items',[]):
      item = itemref();
      if item is not None:
        cost = 0;
        parent = item.parent();
        while parent:
          if not parent.isExpanded():
            cost += 1;
            # if we hit a closed parent tagged with this attr (this is for "All nodes" 
            # and "By class" items), set a cost of None and give up: this item is ineligible
            if getattr(parent,'_no_auto_open',False):
              cost = None;
              break;
          # if parent is a recent item, decrease cost -- we want this item to be preferred
          if parent is self._recent_item:
            _dprint(3,"parent",parent.text(0),"is recent, what a bonus");
            cost -= 100;
          parent = parent.parent();
        _dprint(3,"item cost is ",cost);
        # update best cost estimate
        if cost is not None and cost < best_cost:
          (best_cost,best_item) = (cost,item);
    # if no suitable item was found, try to expand the root tree
    if best_item: 
      _dprint(3,"have a preferred item, ensuring visibility");
    else: # no suitable item was found, try to expand the root tree
      _dprint(3,"no suitable items found, expanding tree to look for active item");
      best_item = self.expand_active_tree(active_ni=node.nodeindex);
    if best_item:
      self._tw.scrollToItem(best_item);
    
  def expand_active_tree (self,start=None,active_ni=None):
    """expands all items from start (the "Root nodes" menu, if start=None) that
    correspond to active (i.e. non-idle) nodes. If 'active' is specified and is a node, 
    returns the item corresponding to that node (if it finds it)."""
    if start is None:
      try: start = self._tw_rootitem;
      except AttributeError: return;
      if not start:
        return None;
    active_item = None;
    start.setExpanded(True);
    # recursively expand all non-idle child nodes
    for i in range(start.childCount()):
      item = start.child(i);
      if isinstance(item,self.NodeItem):
        if item.node.nodeindex == active_ni and not active_item:   
          self._tw.scrollToItem(item);
          active_item = item;
        if (item.node.control_status&meqds.CS_MASK_EXECSTATE) != meqds.CS_ES_IDLE:
          i1 = self.expand_active_tree(item,active_ni);
          active_item = active_item or i1;
    return active_item;
    
  def _view_forest_state (self,viewer=None,**kws):
    _dprint(2,"adding viewer for forest state");
    item = Grid.DataItem('/forest',name='Forest state',caption='<b>Forest state</b>',
                          desc='State of forest',data=meqds.get_forest_state(),
                          refresh=meqds.request_forest_state,viewer=viewer);
    Grid.addDataItem(item,show_gw=True,**kws);
  
  # slot: called to show a context menu for a browser item
  def _item_event (self,item,event,*args,**kw):
    handlers = getattr(item,'_item_event_handler',None);
    if handlers:
      hh = handlers.get(event,None);
      if callable(hh):
        hh(*args,**kw);
  
  def _show_context_menu (self,item,point,col):
    _dprint(3,item,point,col);
    self._item_event(item,'menu',point,col);
      
  def _item_clicked (self,button,item,point,col):
    _dprint(3,button,item,point,col);
    if button == 1 or button == 4:
      self._recent_item = item;
    self._item_event(item,('click',button),point,col);
        
  def _set_recent_item (self,item,previous=None):
    self._recent_item = item;
  
  def _expand_node (self,item):
    self._recent_item = item;
    if isinstance(item,self.NodeItem):
      item.expand();
    else:
      # populate list when first opened, if an iterator is present as an attribute
      iter_nodes = getattr(item,'_iter_nodes',None);
      if iter_nodes:
        busy = BusyIndicator();
        item.addChildren([ self.NodeItem(self,node,node.name) for node in iter_nodes ]);
        delattr(item,'_iter_nodes');
  # _expand_node = busyCursorMethod(_expand_node);
  
  def _request_nodelist (self):
    _dprint(1,"requesting node list and forest state");
    meqds.request_forest_state();
    meqds.request_nodelist(force=True);
    
  def _abort_execute (self):
    if QMessageBox.warning(self.wtop(),"Aborting execution",
          "Abort execution of current request?",
          QMessageBox.Yes,QMessageBox.No|QMessageBox.Default|QMessageBox.Escape) == QMessageBox.No:
      return;
    mqs().meq('Execute.Abort',record(),wait=False);
    
  def _debug_set_verbosity_slot (self,qa):
    """Sends a debug level message.""";
    if not getattr(self,'_setting_debug_control',False):
      mqs().meq('Debug.Set.Level',record(debug_level=qa._debug_level,get_forest_status=2),wait=False);

  def _debug_single_step (self):
    mqs().meq('Debug.Single.Step',record(),wait=False);
  
  def _debug_next_node (self):
    mqs().meq('Debug.Next.Node',record(),wait=False);
    
  def _debug_until_node (self,node):
    mqs().meq('Debug.Until.Node',record(nodeindex=node.nodeindex),wait=False);
  
  def _debug_continue (self):
    mqs().meq('Debug.Continue',record(),wait=False);
    
  def _debug_pause (self):
    mqs().meq('Debug.Interrupt',record(),wait=False);
    
  def _debug_bp_on_fail (self,on):
    if getattr(self,'_setting_debug_control',False):
      return;
    if on:
      mqs().meq('Set.Forest.Breakpoint',record(breakpoint=meqds.BP_FAIL),wait=False);
    else:
      mqs().meq('Clear.Forest.Breakpoint',record(breakpoint=meqds.BP_FAIL),wait=False);

  def add_action (self,action,order=1000,where="toolbar",callback=None):
    if callback:
      QObject.connect(action,SIGNAL("triggered(bool)"),callback);
    self._actions.setdefault(where,[]).append((order,action));
    
  def add_separator (self,order=1000,where="toolbar"):
    self._actions.setdefault(where,[]).append((order,None));
    
  def add_stretch (self,order=1000,where="toolbar"):
    self._actions.setdefault(where,[]).append((order,"stretch"));
    
  def get_action_list (self,which):
    return self._actions.get(which,[]);

  #def save_forest_dialog (self):
    #try: dialog = self._save_forest_dialog;
    #except AttributeError:
      #self._save_forest_dialog = dialog = QFileDialog(self.wtop(),"save dialog",True);
      #dialog.resize(800,500);
      #dialog.setMode(QFileDialog.AnyFile);
      #dialog.setFilters("Forests (*.forest *.meqforest);;All files (*.*)");
      #dialog.setViewMode(QFileDialog.Detail);
      #dialog.setCaption("Save forest");
    #else:
      #dialog.rereadDir();
    #if dialog.exec_loop() == QDialog.Accepted:
      #fname = str(dialog.selectedFile());
      #rec = record(file_name=fname,get_forest_status=True);
      #mqs().meq('Save.Forest',rec,wait=False);

  #def load_forest_dialog (self):
    #try: dialog = self._load_forest_dialog;
    #except AttributeError:
      #self._load_forest_dialog = dialog = QFileDialog(self.wtop(),"load dialog",True);
      #dialog.resize(800,500);
      #dialog.setMode(QFileDialog.ExistingFile);
      #dialog.setFilters("Forests (*.forest *.meqforest);;All files (*.*)");
      #dialog.setViewMode(QFileDialog.Detail);
      #dialog.setCaption("Load forest");
    #else:
      #dialog.rereadDir();
    #if dialog.exec_loop() == QDialog.Accepted:
      #fname = str(dialog.selectedFile());
      #rec = record(file_name=fname,get_forest_status=True);
      #meqds.clear_forest();
      #mqs().meq('Load.Forest',rec,wait=False);
      #self._request_nodelist();
      
  def show_icon_reference (self):
    try: iconref = self._icon_reference;
    except AttributeError:
      iconref = self._icon_reference = NodeIconReference(self.wtop().topLevelWidget());
    iconref.setVisible(not iconref.isVisible());
      
class NodeIconReference (QMainWindow):
  RefList = None;
  def __init__ (self,parent):
    QMainWindow.__init__(self,parent,Qt.Dialog|Qt.WindowStaysOnTopHint);
    self.setWindowTitle("Node Icon Reference");
    self.setWindowIcon(pixmaps.info_blue_round.icon());
    # define list
    if not self.RefList:
      self.RefList = [ \
        (None,None) ];
      # add state icons
      self.RefList += [ (pixmap,desc) for (val,name,char,desc,pixmap) in meqds.CS_ES_statelist ];
      # add result icons
      self.RefList += [ \
        (None,None),\
        (None,"      Node result icons:"),
        (None,None),\
        (None,"no node result (no icon)"),
        (pixmaps.blue_round_reload,"new result, cached"),
        (pixmaps.blue_round_empty,"new empty result, cached"),
        (pixmaps.green_return,"new result, not cached"),
        (pixmaps.blue_round_reload_return,"reused result from cache"),
        (pixmaps.blue_round_empty_return,"reused empty result from cache"),
        (pixmaps.blue_round_clock,"'wait' result"),
        (pixmaps.grey_round_cross,"'missing' result"),
        (pixmaps.red_round_cross,"'fail' result"),
        (None,None),\
        (None,"      Miscellaneous icons:"),
        (None,None),\
        (pixmaps.clear_red_right,"node is stopped in debugger"),
        (pixmaps.breakpoint,"node has breakpoints set"),
        (pixmaps.forward_to,"node has a temp ('until') breakpoint set"),
        (pixmaps.publish,"node is publishing snapshots"),
        (pixmaps.publish_active,"node state is updated (brief flash)"),
        (pixmaps.cancel,"node is disabled") ];
    #
    wtop = QWidget(self);
    lo = QGridLayout(wtop);
    lo.setSpacing(3);
    self.setCentralWidget(wtop);
    for row,(pixmap,text) in enumerate(self.RefList):
      if text is None:  # draw separator 
        icon = QWidget(wtop); 
        label = QFrame(wtop);
        label.setFrameShape(QFrame.HLine);
        label.setFrameShadow(QFrame.Sunken);
      else: # draw entry
        if pixmap is None:
          icon = QWidget(wtop);
        else:
          icon = QLabel(wtop);
          icon.setPixmap(pixmap.pm());
        label = QLabel(text,wtop);  
      # make icon non-resizable
      icon.setSizePolicy(QSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum));
      # add to layout
      lo.addWidget(icon,row,0);
      lo.addWidget(label,row,1);
    wtop.setSizePolicy(QSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum));
    self.setSizePolicy(QSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum));
    
  def mouseReleaseEvent (self,ev):
    self.hide();
    
def define_treebrowser_actions (tb):
  _dprint(1,'defining standard treebrowser actions');
  parent = tb.wtop();
  # populate the toolbar
  # Refresh
  refresh = tb._qa_refresh = QAction(pixmaps.refresh.icon(),"Refresh forest",parent); 
  refresh.setShortcut(Qt.Key_F2);
  refresh.setToolTip("This asks for an update of the current forest from the meqserver");
  refresh._is_enabled = lambda tb=tb: tb.is_connected;
  tb.add_action(refresh,10,callback=tb._request_nodelist);
  tb.add_separator(20);
  ## Save and load
  #qa_load = tb._qa_load = QAction(pixmaps.file_open.icon(),"Load forest",0,parent);
  #qa_save = tb._qa_save = QAction(pixmaps.file_save.icon(),"Save forest",0,parent);
  #qa_save.setToolTip("This saves the current forest to a file");
  #qa_load.setToolTip("This loads the current forest from a file");
  #qa_load._is_enabled = qa_save._is_enabled = lambda tb=tb: tb.is_connected and not tb.is_running;
  #tb.add_action(qa_load,30,callback=tb.load_forest_dialog);
  #tb.add_action(qa_save,40,callback=tb.save_forest_dialog);
  #tb.add_separator(45);
  # Abort button
  qa_abort = QAction(pixmaps.abort.icon(),"&Abort execution",parent);
  qa_abort.setToolTip("This aborts the currently executing request");
  qa_abort._is_visible = lambda tb=tb: tb.is_connected;
  qa_abort._is_enabled = lambda tb=tb: tb.is_running;
  tb.add_action(qa_abort,50,callback=tb._abort_execute);
  tb.add_separator(55);
  # debugger verbosity
  # dbg_enable = tb._qa_dbg_enable = QAction(pixmaps.debugger.icon(),"Enable verbose &debugger",Qt.Key_F5,parent);
  dbg_verbosity = tb._qa_dbg_verbosity = QActionGroup(parent);
  dbg_verbosity.setExclusive(True);
  # dbg_verbosity.setUsesDropDown(True);
  QObject.connect(dbg_verbosity,SIGNAL("triggered(QAction*)"),tb._debug_set_verbosity_slot);
  # dbg_verbosity.setToolTip("""This changes debug verbosity levels""");
  tb._qa_dbg_verbosities = [];
  for level,pm,text in \
     [ (0,pixmaps.debug0,"Verbosity level 0: fast, no node tracking"), 
       (1,pixmaps.debug1,"Verbosity level 1: slow, keep track of node result codes"), 
       (2,pixmaps.debug2,"Verbosity level 2: slowest, keep track of everything") ]:
    qa = QAction(pm.icon(),text,dbg_verbosity);
    qa.setCheckable(True);
    qa._debug_level = level;
    qa._is_visible = lambda tb=tb: tb.is_connected;
    qa._is_enabled = lambda tb=tb: tb.is_connected;
    tb._qa_dbg_verbosities.append(qa);
    tb.add_action(qa,60);
  tb.add_separator(61);
  # Breakpoint on fail
  bp_on_fail = tb._qa_bp_on_fail = QAction(pixmaps.breakpoint_on_fail.icon(),"Set global breakpoint on &fail",parent);
  bp_on_fail.setCheckable(True);
  bp_on_fail.setToolTip("This enables or disables a global breakpoint on any node returning a FAIL result");
  bp_on_fail._is_visible = lambda tb=tb: tb.is_connected;
  bp_on_fail._is_enabled = lambda tb=tb: tb.is_connected;
  QObject.connect(bp_on_fail,SIGNAL("triggered(bool)"),tb._debug_bp_on_fail);
  tb.add_action(bp_on_fail,65);
  tb.add_separator(66);
  # Pause
  pause = QAction(pixmaps.pause.icon(),"&Interrupt",parent);
  pause.setShortcut(Qt.Key_F6);
  pause._is_visible = lambda tb=tb: tb.is_connected;
  pause._is_enabled = lambda tb=tb: tb.is_running and not tb.is_stopped;
  tb.add_action(pause,70,callback=tb._debug_pause);
  tb.add_separator(80);
  # Debug action group
  ag_debug = tb._qa_dbg_tools = QActionGroup(parent);
  ag_debug.setExclusive(False);
  dbgcont  = QAction(pixmaps.right_2triangles.icon(),"&Continue",ag_debug);
  dbgcont.setShortcut(Qt.Key_F6+Qt.SHIFT);
  tb.add_action(dbgcont,90);
  QObject.connect(dbgcont,SIGNAL("triggered()"),tb._debug_continue);
  dbgstep  = QAction(pixmaps.down_triangle.icon(),"&Step",ag_debug);      
  dbgstep.setShortcut(Qt.Key_F7);
  tb.add_action(dbgstep,91);
  QObject.connect(dbgstep,SIGNAL("triggered()"),tb._debug_single_step);
  dbgnext  = QAction(pixmaps.down_2triangles.icon(),"Step to &next node",ag_debug);      
  dbgnext.setShortcut(Qt.Key_F8);
  tb.add_action(dbgnext,92);
  QObject.connect(dbgnext,SIGNAL("triggered()"),tb._debug_next_node);
  for qa in ag_debug.actions():
    qa._is_visible = lambda tb=tb: tb.is_connected;
    qa._is_enabled = lambda tb=tb: tb.is_loaded and tb.is_stopped;
  
  # show node icon reference
  tb.add_stretch(1000);
  show_help = QAction(pixmaps.info_blue_round.icon(),"Show icon &reference...",parent);
  show_help.setShortcut(Qt.CTRL+Qt.Key_F1);
  QObject.connect(show_help,SIGNAL("triggered()"),tb.show_icon_reference);
  show_help._is_enabled = lambda tb=tb:True;
  tb.add_action(show_help,1010);
  tb.add_action(show_help,5,where="node");
  
  # populate node context menu
  tb.add_action(NA_NodeDisable,10,where="node");
  tb.add_action(NA_NodePublish,20,where="node");

  # populate debug context sub-menu
  tb.add_action(NA_ContinueUntil,10,where="debug");
  tb.add_action(dbgcont,20,where="debug");
  tb.add_action(dbgstep,21,where="debug");
  tb.add_action(dbgnext,22,where="debug");

class NodeAction (object):
  """NodeAction is a class implementing a node-associated action.""";
  # static attributes of action
  # these are meant to be redefined by subclasses
  text = 'unknown action';
  icon = None;
  nodeclass = None;  # None means all nodes; else assign meqds.NodeClass('class');
  # static methods describing properties of this action
  def applies_to_node (self,node):
    _dprint(3,'nodeclass',self.nodeclass,'applies to',node,'?');
    if self.nodeclass is None:
      return True;
    if isinstance(self.nodeclass,str):
      self.nodeclass = meqds.NodeClass(self.nodeclass);
    _dprint(3,self.nodeclass,meqds.NodeClass(node.classname));
    try: applies = issubclass(meqds.NodeClass(node.classname),self.nodeclass);
    except: applies = False;
    _dprint(3,'applies:',applies);
    return applies;
  applies_to_node = classmethod(applies_to_node);
  
  def __init__ (self,item,menu,callback=None,separator=False):
    """instantiates action and adds it to the given menu.
    item is a TreeBrowser.NodeItem with which the menu is associated."""
    self.item = weakref_proxy(item);
    self.node = weakref_proxy(item.node);
    self.tb   = weakref_proxy(item.tb);
    if separator:
      menu.addSeparator();
    self._callback = callback or xcurry(self.activate,_args=(self.node,));
    if self.icon:
      self.item_qa = menu.addAction(self.icon(),self.text,self._callback);
    else:
      self.item_qa = menu.addAction(self.text,self._callback);
      
  def is_checked (self,node):
    return None;
    
  def is_enabled (self,node):
    return True;
      
  def update (self,menu,node):
    """this is called whenever a menu is being displayed. Subclasses may
    redefine this to set up the item as appropriate
    The node argument defines which node the action is being applied to, or None 
    if the action is global. For actions residing in context menus, node is the 
    same as self.node.""";
    # enable/disable item
    self.item_qa.setEnabled(self.is_enabled(node));
    # if is_checked returns not None, set item as checked/unchecked
    checked = self.is_checked(node);
    if checked is not None:
      self.item_qa.setCheckable(True);
      self.item_qa.setChecked(checked);
      
  def activate (self,node):
    """callback: called whenever the menu item is selected. 
    The node argument defines which node the action is being applied to, or None 
    if the action is global. For actions residing in context menus, node is the 
    same as self.node.""";
    pass;
      
class NA_NodeDisable (NodeAction):
  text = "Disable";
  icon = pixmaps.cancel.icon;
  def activate (self,node):
    cs = node.control_status ^ meqds.CS_ACTIVE;
    meqds.set_node_state(node,control_status=cs);
  # define is_checked as a property that is computed on-the-fly
  def is_checked (self,node):
    return not node.is_active();

class NA_NodePublish (NodeAction):
  text = "Publish";
  icon = pixmaps.publish.icon;
  def activate (self,node):
    meqds.enable_node_publish(node,not node.is_publishing());
  def is_checked (self,node):
    return node.is_publishing();

class NA_ContinueUntil (NodeAction):
  text = "Continue until";
  icon = pixmaps.forward_to.icon;
  def __init__ (self,item,menu,**kw):
    NodeAction.__init__(self,item,menu,**kw);
    menu.changeItem(self.item_id,"Continue &until "+item.node.name);
  def activate (self,node):
    self.tb._debug_until_node(node);
  def is_enabled (self,node):
    return self.tb.is_running;



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from qwt.qt.QtGui import QApplication,QHBoxLayout, QLabel, QSizePolicy, QSpacerItem
from qwt.qt.QtGui import QWidget
from qwt.qt.QtCore import Qt

import numpy
from math import sin
from math import cos
from math import pow
from math import sqrt

# modules that are imported
HAS_TIMBA = False
try:
  from Timba.dmi import *
  from Timba import utils
  from Timba.Meq import meqds
  from Timba.Meq.meqds import mqs
  from MeqGUI.GUI.pixmaps import pixmaps
  from MeqGUI.GUI import widgets
  from MeqGUI.GUI.browsers import *
  from MeqGUI import Grid
  from MeqGUI.Plugins.VellsData import VellsData
  from MeqGUI.Plugins.ResultsRange_qt5 import ResultsRange
  from MeqGUI.Plugins.DataDisplayMainWindow_qt5 import DisplayMainWindow

  from Timba.utils import verbosity
  _dbg = verbosity(0,name='collections_plotter');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
 import traceback
 traceback.print_exc()
 print("Cannot import Plotting and GUI plugins... plotting will not work!!")



class CollectionsPlotter(GriddedPlugin):
  """ a class to visualize data, VellSets or visu data, that is 
      contained within a node's cache_result record. Objects of 
      this class are launched from the meqbrowser GUI """

  _icon = pixmaps.bars3d
  viewer_name = "Collections Plotter";
  def is_viewable (data):
    return len(data) > 0;
  is_viewable = staticmethod(is_viewable);

  def __init__(self,gw,dataitem,cellspec={},**opts):
    GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    """ Instantiate plotter object that is used to display plots
    """
    self.set_init_parameters()
# back to 'real' work
    self.dataitem = dataitem
    if dataitem and dataitem.data is not None:
      self.set_data(dataitem);

  def set_init_parameters(self):
    self._rec = None;
    self._plot_type = None
    self._wtop = None;
    self._vells_data = None
    self.layout_created = False
    self.layout_parent = None
    self.layout = None
    self._visu_plotter = None
    self.counter = 0
    self.max_range = 0
    self._node_name = None
    self._prev_rq_id = -1
    self._plot_label = None
    self._tab_label = 'Page'
    self.prev_label = "===="

  def wtop (self):
    """ function needed by Oleg for reasons known only to him! """
    return self._wtop;

  def create_layout_stuff(self):
    """ create grid layouts into which plotter widgets are inserted """
    if self.layout_parent is None or not self.layout_created:
      self.layout_parent = QWidget(self.wparent())
      self.layout = QHBoxLayout(self.layout_parent)
      self.set_widgets(self.layout_parent,self.dataitem.caption,icon=self.icon())
      self.layout_created = True
    self._wtop = self.layout_parent;       
    self.create_2D_plotter()

  
  def create_2D_plotter(self):
    if self._visu_plotter is None:
      self._visu_plotter = DisplayMainWindow(parent=None,name=" ", num_curves=self._max_per_display, plot_label=self._plot_label)
      self.layout.addWidget(self._visu_plotter,100)
      self._visu_plotter.showMessage.connect(self._gw.show_message)
      lo0 = QVBoxLayout();
      lo0.setContentsMargins(0,0,0,0);
      self.layout.addLayout(lo0,0);

      self._results_range = ResultsRange(parent=self.layout_parent, name="", horizontal=False,add_spacer=False)
      self._results_range.setMaxValue(max_value= 0, allow_shrink=True)
      lo0.addWidget(self._results_range)
      self._tab_selector = ResultsRange(parent=self.layout_parent,horizontal=True, hide_slider=True,add_spacer=False)
      self._tab_selector.setLabel('Page selector:')
      self._tab_selector.hideNDControllerOption()
      self._tab_selector.set_offset_index(0)
      self._tab_selector.set_emit(True)
      lo0.addWidget(self._tab_selector);

      # add data selectors 
      self._visu_plotter.createDataSelectorWidgets(self.layout_parent,lo0);
      spacer = QSpacerItem(9,22,QSizePolicy.Minimum,QSizePolicy.Expanding)
      lo0.addItem(spacer)
       
      self._visu_plotter.show()
      self._results_range.set_offset_index(0)
      self._visu_plotter.auto_offset_value.connect(self.set_range_selector)
      self._visu_plotter.number_of_tabs.connect(self.updateTabSelectorRange)
      self._results_range.result_index.connect(self._visu_plotter.set_range_selector)
      self._results_range.set_auto_scaling.connect(self._visu_plotter.set_auto_scaling)
      self._tab_selector.result_index.connect(self._visu_plotter.change_tab_page)
  # create_2D_plotter

  def set_range_selector(self, max_range):
    """ set or update maximum range for slider controller """
    self.max_range = max_range
    self._results_range.set_emit(False)
    self._results_range.setMinValue(0)
    self._results_range.setMaxValue(max_range)
    self._results_range.setTickInterval( max_range // 10 )
    self._results_range.setValue(max_range)
    self._results_range.setLabel('offset:',align= Qt.AlignHCenter)
    self._results_range.hideNDControllerOption()
    self._results_range.set_emit(True)
    self._results_range.show()

  def updateTabSelectorRange(self, max_range):
    """ set or update maximum range for tab page selector """
    self._tab_selector.set_emit(False)
    self._tab_selector.setMinValue(0)
    self._tab_selector.setMaxValue(max_range,allow_shrink=True)
    self._tab_selector.set_emit(True)


  def set_data (self,dataitem,default_open=None,**opts):
    """ this callback receives data from the meqbrowser, when the
        user has requested a plot. It decides whether the data is
        from a VellSet or visu data record, and  after any
        necessary preprocssing forwards the data to one of
        the functions which does the actual plotting """

    if HAS_TIMBA: _dprint(3, '** in collections_plotter:set_data callback')
    self._rec = dataitem.data;
    if HAS_TIMBA:
      _dprint(3, 'set_data: initial self._rec ', self._rec)
      _dprint(3, 'incoming record has keys ', list(self._rec.keys()))
# if we are single stepping through requests, Oleg may reset the
# cache, so check for a non-data record situation
    if self._rec is None:
      return
    if isinstance(self._rec, bool):
      return

    try:
      self._node_name = self._rec.name
    except:
      pass
    if HAS_TIMBA:
      _dprint(3, 'node name is ', self._node_name)

    if hasattr(self._rec,'plot_label'):
      self._plot_label = self._rec.plot_label;
    if hasattr(self._rec,'tab_label'):
      self._tab_label = self._rec.tab_label;
    self._vells_label = getattr(self._rec,'vells_label',[]);
    if not isinstance(self._vells_label,(list,tuple)):
      self._vells_label = [];
# previous_statement_gives example output: self._vells_label ('XX', 'XY', 'YX', 'YY'
    self.label = '';  # extra label, filled in if possible
# there's a problem here somewhere ...
    if dmi_typename(self._rec) != 'MeqResult': # data is not already a result?
      # try to put request ID in label
      rq_id_found = False
      data_failure = False
      try:
        if "request_id" in self._rec.cache:
          self.label = "rq " + str(self._rec.cache.request_id);
          rq_id_found = True
        if "result" in self._rec.cache:
          self._rec = self._rec.cache.result; # look for cache.result record
          if not rq_id_found and "request_id" in self._rec:
            self.label = "rq " + str(self._rec.request_id);
            rq_id_found = True
        else:
          data_failure = True
        if HAS_TIMBA: _dprint(3, 'we have req id ', self.label)
      except:
        data_failure = True
      if data_failure:
        if HAS_TIMBA: _dprint(3, ' we have a data failure')
# cached_result not found, display an empty viewer with a "no result
# in this node record" message (the user can then use the Display with
# menu to switch to a different viewer)
        Message = "No cache result record was found for this node, so no plot can be made."
        cache_message = QLabel(Message,self.wparent())
        cache_message.setTextFormat(Qt.RichText)
        self._wtop = cache_message
        self.set_widgets(cache_message)
        self.set_init_parameters()
        return
      if HAS_TIMBA: _dprint(3, 'collections: request id ', self.label)
      end_str = len(self.label)
      self._rq_id_end = int(self.label[end_str-1:end_str])
      if self._rq_id_end == self._prev_rq_id:
        new_plot = False
      else:
        new_plot = True
        self._prev_rq_id = self._rq_id_end
      
      if self.label.find(self.prev_label) >= 0:
# we have already plotted this stuff, so return
        return
      else:
        self.prev_label = self.label

# are we dealing with Vellsets?
    self.counter = self.counter + 1
    self._max_per_display = 64
    if "dims" in self._rec:
      if HAS_TIMBA: _dprint(3, '*** dims field exists ', self._rec.dims)
      self.dims = list(self._rec.dims)
    else:
      self.dims = None
    if "vellsets" in self._rec:
      self._number_of_planes = len(self._rec["vellsets"])
      self.dims_per_group = 1
      index_labels = None;
      # make labels for multi-dim result
      if not self.dims is None:
        # in collections plotter, assume that first 'dims' 
        # corresponds to tracks that are displayed , so determine
        # size of group from second dimension and up
        if len(self.dims) > 1:
          dims_start = 1
          for i in range(dims_start,len(self.dims)):
            self.dims_per_group = self.dims_per_group * self.dims[i]
          if self.dims_per_group == len(self._vells_label):
            index_labels = self._vells_label;
          else:
            if len(self._vells_label) > 0:
              if self._vells_label[0] == 'XX' or self._vells_label[0] == 'RR':
              # we have correlator data to display 
              # - ignore self.dims_per_group stuff
                index_labels = self._vells_label;
            else:
            # setup index array -- this makes a list such as [1,1],[1,2],[2,1],[2,2] (for e.g. a 2x2 array)
              indices = numpy.ndindex(*self.dims[dims_start:]);
            # sep = "," if any([dim>9 for dim in self.dims[dims_start:]]) else "";
              index_labels = [ ",".join([str(x+1) for x in ind]) for ind in indices ];
      # index labels not set by the above? set them here
      if index_labels is None:
        if self._number_of_planes == len(self._vells_label):
          index_labels = self._vells_label;
        else:
          index_labels = list(range(0,1));
      # replace index labels with vells labels
      if self._visu_plotter is None:
        self.create_layout_stuff()
      if new_plot: 
        self._visu_plotter.setNewPlot()
      data_dict = {}
      self._visu_plotter.setDataElementLabels(index_labels,list(self.dims[1:]) if self.dims is not None else (1,));
      for i in range(self._number_of_planes):
        channel = int(i / self.dims_per_group)
        if self._node_name is None:
          data_dict['source'] = "Channel " + str(channel)
        else:
          data_dict['source'] = self._node_name
        data_dict['channel'] = channel
        data_dict['sequence_number'] = self.counter
        screen_num = channel // self._max_per_display
#       data_dict['data_type'] = self._tab_label + ' ' + str(screen_num+1)
        data_dict['data_type'] = self._tab_label + ' ' + str(screen_num)
        index = i - channel * self.dims_per_group
        
        if index == 0:
          data_dict['value'] = {}
          data_dict['flags'] = {}
        try:
          datakey = index_labels[index];
        except:
          pass
#       if channel == 100:
#         data_dict['value'][index] = 100.0 * self._rec.vellsets[i].value
#       else:

# hopefully handle cases with non-existent results
        try:
          if  "value" in self._rec.vellsets[i]:
            data_dict['value'][datakey] = self._rec.vellsets[i].value
          else:
            data_dict['value'][datakey] = None
        except:
            data_dict['value'][datakey] = None
# for test purposes
#       if index == 2:
#           data_dict['value'][datakey] = None
        try:
          if  "flags" in self._rec.vellsets[i]:
            data_dict['flags'][datakey] = self._rec.vellsets[i].flags
          else:
            data_dict['flags'][datakey] = None
        except:
            data_dict['flags'][datakey] = None
        if index == self.dims_per_group-1:
          self._visu_plotter.updateEvent(data_dict)
          data_dict = {}

# enable & highlight the cell
    self.enable();
    self.flash_refresh();
    if HAS_TIMBA: _dprint(3, 'exiting process_data')

Grid.Services.registerViewer(meqds.NodeClass(),CollectionsPlotter,priority=30)




#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

# #from . import array_browser
# from . import array_plotter
# from . import BufferSizeDialog_qt5
# from . import chartplot_qt5
# from . import collections_plotter
# from . import ComplexScaleDraw_qt5
# from . import DataDisplayMainWindow_qt5
# #from . import dialog_window_config
# from . import display_image_qt5
# #from . import guiplot2dnodesettings
# from . import ImageScaler
# from . import ND_Controller_qt5
# from . import node_execute
# from . import plot_printer_qt5
# from . import printfilter_qt5
# from . import pylab_plotter
# from . import quickref_plotter
# from . import QwtColorBar_qt5
# from . import QwtPlotImage_qt5
# from . import QwtSpy_qt5
# from . import result_plotter
# from . import ResultsRange_qt5
# from . import SolverData
# from . import SpectrumData
# from . import TableInspector
# from . import VellsData
# from . import VellsTree_qt5
# from . import zoomwin_qt5


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.dmi import *
from Timba.utils import *

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

from MeqGUI import Grid
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI import meqgui

_dbg = verbosity(0,name='bookmarks');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

class Bookmark (QObject):
  def __init__ (self,name,udi,parent,viewer=None):
    QObject.__init__(self,parent);
    QObject.connect(self,PYSIGNAL("updated()"),parent,PYSIGNAL("updated()"));
    self.name = name;
    self.rec = record(name=name,udi=udi);
    # resolve viewer to actual viewer class (or None if none specified),
    # and rec.viewer to viewer name; set icon based on viewer
    if viewer is None:
      self.icon = QIcon;
      self.enabled = True;
      self.viewer  = None;
    else:
      if isinstance(viewer,str):
        self.rec.viewer = viewer;
        self.viewer = Grid.Services.getViewerByName(viewer);
      elif callable(viewer):
        self.rec.viewer = getattr(viewer,'viewer_name',viewer.__name__);
        self.viewer = viewer;
      else:
        raise TypeError("illegal viewer argument "+str(viewer));
      if self.viewer is not None:
        self.icon = getattr(self.viewer,'icon',QIcon);
        self.enabled = True;
      else:
        self.icon = QIcon;
        self.enabled = False;
  def show(self,**kws):
    self.parent().emit(SIGNAL("showBookmark"),self.rec.udi,self.rec.viewer);

class Pagemark (QObject):
  icon = pixmaps.bookmark_toolbar.icon;
  enabled = True;
  def __init__ (self,name,page,parent):
    QObject.__init__(self,parent);
    QObject.connect(self,PYSIGNAL("updated()"),parent,PYSIGNAL("updated()"));
    self.name = name;
    self.page = page;
    self.rec = record(name=name,page=page);
  def show (self,**kws):
    self.parent().emit(SIGNAL("showPagemark"),self,);

class BookmarkFolder (QObject):
  icon = pixmaps.bookmark_folder.icon;
  enabled = True;
  def __init__ (self,name,parent,menu=None,load=None,gui_parent=None):
    """Creates a folder of bookmarks, and the associated menu.
    If 'menu' is not None, bookmarks will be added to the supplied menu, else a new one will be created.
    If load is set to a list (of bookmarks records), folder will be loaded.
    'parent' is parent object. 'gui_parent' is parent object for menus (or else 'parent' will be used.)
    """;
    QObject.__init__(self,parent);
    self.name = name;
    self._gui_parent = gui_parent;
    self._bklist = [];
    self._qas = [];
    if menu:
      self._menu = menu;
      self._have_initial_menu = True;
    else:
      self._menu = QMenu(str(name),gui_parent or parent);
      self._menu.setIcon(self.icon());
      self._have_initial_menu = False;
    if load is not None:
      self.load(load);
    QObject.connect(self,PYSIGNAL("showBookmark()"),self.parent(),PYSIGNAL("showBookmark()"));
    QObject.connect(self,PYSIGNAL("showPagemark()"),self.parent(),PYSIGNAL("showPagemark()"));
    
  def load (self,bklist):
    """loads bookmarks from list""";
    self._bklist = [];
    if self._have_initial_menu:
      for qa in self._qas:
        self._menu.removeAction(qa);
    else:
      self._menu.clear();
    self._qas = [];
    for item in bklist:
      try: name = item.name;
      except AttributeError:
        _dprintf(2,"ignoring invalid bookmark item",item);
        continue;
      # add all items here too!
      # determine type
      if hasattr(item,'page'):  # page bookmark
        self.addItem(Pagemark(name,item.page,self),quiet=True);
      elif hasattr(item,'folder'): # folder
        self.addItem(BookmarkFolder(name,self,load=item.folder,gui_parent=self._gui_parent),quiet=True);
      elif hasattr(item,'udi'): # bookmark
        self.addItem(Bookmark(name,item.udi,self,viewer=getattr(item,'viewer',None)),quiet=True);
      else:
        _dprintf(2,"ignoring invalid bookmark item",item);
    
  def getList (self):
    """returns bookmarks as a list""";
    outlist = [];
    for item in self._bklist:
      if isinstance(item,BookmarkFolder):
        outlist.append(record(name=item.name,folder=item.getList()));
      elif isinstance(item,Bookmark):
        outlist.append(item.rec);
      elif isinstance(item,Pagemark):
        outlist.append(item.rec);
    return outlist;
    
  def addItem (self,item,quiet=False):
    """adds an item to the list and menu.""";
    self._bklist.append(item);
    if isinstance(item,BookmarkFolder):
      qa = self._menu.addMenu(item.getMenu());
    elif isinstance(item,Pagemark) or isinstance(item,Bookmark):
      qa = self._menu.addAction(item.icon(),str(item.name),item.show);
    if not item.enabled:
      qa.setEnabled(False);
    self._qas.append(qa);
    if not quiet:
      self.emit(SIGNAL("updated"));
    return item;
    
  def add (self,name,udi,viewer=None):
    """adds bookmark to end of list, updates popup menu if present.""";
    return self.addItem(Bookmark(name,udi,self,viewer=viewer));
      
  def addFolder (self,name):
    """adds bookmark folder to end of list, updates popup menu if present.""";
    return self.addItem(BookmarkFolder(name,self));
  
  def addPage (self,name,page):
    """adds page bookmark to end of list, updates popup menu if present.""";
    return self.addItem(Pagemark(name,page,self));
    
  def getMenu (self):
    """returns popup menu containing all items.""";
    return self._menu;
    
  def generatePageName (self):
    """convenience function to generate a default name for a new pagemark.""";
    # use list comprehension to count the number of pagemarks
    count = len([item for item in self._bklist if isinstance(item,Pagemark)]);
    return "Page " + str(count+1);
    
